# #############################################################################
# recon.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Reconstruction
==============

Check out `this notebook <https://drive.google.com/file/d/1Wgt6ZMRZVuctLHaXxk7PEyPaBaUPvU33/view?usp=drive_link>`_
on Google Colab for an overview of the reconstruction algorithms available in LenslessPiCam (analytic and learned).

The core algorithmic component of ``LenslessPiCam`` is the abstract
class :py:class:`~lensless.ReconstructionAlgorithm`. The five reconstruction
strategies available in ``LenslessPiCam`` derive from this class:

-  :py:class:`~lensless.GradientDescent`: projected gradient descent with a
   non-negativity constraint. Two accelerated approaches are also
   available: :py:class:`~lensless.NesterovGradientDescent` and
   :py:class:`~lensless.FISTA`.
-  :py:class:`~lensless.ADMM`: alternating direction method of multipliers (ADMM)
   with a non-negativity constraint and a total variation (TV)
   regularizer [1]_.
-  :py:class:`~lensless.APGD`: accelerated proximal gradient descent with Pycsou
   as a backend. Any differentiable or proximal operator can be used as
   long as it is compatible with Pycsou, namely derives from one of
   `DiffFunc <https://github.com/matthieumeo/pycsou/blob/a74b714192821501371c89dbd44eac15a5456a0f/src/pycsou/abc/operator.py#L980>`_
   or `ProxFunc <https://github.com/matthieumeo/pycsou/blob/a74b714192821501371c89dbd44eac15a5456a0f/src/pycsou/abc/operator.py#L741>`_.
-  :py:class:`~lensless.UnrolledFISTA`: unrolled FISTA with a non-negativity constraint.
-  :py:class:`~lensless.UnrolledADMM`: unrolled ADMM with a non-negativity constraint and a total variation (TV) regularizer [1]_.

Note that the unrolled algorithms derive from the abstract class
:py:class:`~lensless.TrainableReconstructionAlgorithm`, which itself derives from
:py:class:`~lensless.ReconstructionAlgorithm` while adding functionality
for training on batches and adding trainable pre- and post-processing
blocks.

New reconstruction algorithms can be conveniently implemented by
deriving from the abstract class and defining the following abstract
methods:

-  the update step: :py:class:`~lensless.ReconstructionAlgorithm._update`.
-  a method to reset state variables: :py:class:`~lensless.ReconstructionAlgorithm.reset`.
-  an image formation method: :py:class:`~lensless.ReconstructionAlgorithm._form_image`.

One advantage of deriving from :py:class:`~lensless.ReconstructionAlgorithm` is
that functionality for iterating, saving, and visualization is already
implemented. Consequently, using a reconstruction algorithm that derives
from it boils down to three steps:

1. Creating an instance of the reconstruction algorithm.
2. Setting the data.
3. Applying the algorithm.


2D example (ADMM)
-----------------

For example, for ADMM:

.. code:: python

       recon = ADMM(psf)
       recon.set_data(data)
       res = recon.apply(n_iter=n_iter)

A full running example can found in ``scripts/recon/admm.py`` and run as:

.. code:: bash

    python scripts/recon/admm.py

Note that a YAML configuration script is defined in ``configs/admm_thumbs_up.yaml``,
which is used by default. Individual parameters can be configured as such:

.. code:: bash

    python scripts/recon/admm.py admm.n_iter=10 preprocess.gray=True

``--help`` can be used to view all available parameters.

.. code::

    >> python scripts/recon/admm.py --help

    ...

    == Config ==
    Override anything in the config (foo.bar=value)

    files:
        psf: data/psf/tape_rgb.png
        data: data/raw_data/thumbs_up_rgb.png
    preprocess:
        downsample: 4
        shape: null
        flip: false
        bayer: false
        blue_gain: null
        red_gain: null
        single_psf: false
        gray: false
    display:
        disp: 1
        plot: true
        gamma: null
    save: false
    admm:
        n_iter: 5
        mu1: 1.0e-06
        mu2: 1.0e-05
        mu3: 4.0e-05
        tau: 0.0001

    ...


Alternatively, a new configuration file can be defined in the ``configs`` folder and
passed to the script:

.. code:: bash

    python scripts/recon/admm.py -cn <CONFIG_FILENAME_WITHOUT_YAML_EXT>


3D example
----------

It is also possible to reconstruct 3D scenes using :py:class:`~lensless.GradientDescent` or :py:class:`~lensless.APGD`. :py:class:`~lensless.ADMM` does not support 3D reconstruction yet.
This requires to use a 3D PSF as an input in the form of an ``.npy`` or ``.npz`` file, which is a set of 2D PSFs corresponding to the same diffuser sampled with light sources at different depths.
The input data for 3D reconstructions is still a 2D image, as collected by the camera. The reconstruction will be able to separate which part of the lensless data corresponds to which 2D PSF,
and therefore to which depth, effectively generating a 3D reconstruction, which will be outputed in the form of an ``.npy`` file. A 2D projection on the depth axis is also displayed to the user.

The same scripts for 2D reconstruction can be used for 3D reconstruction, namely ``scripts/recon/gradient_descent.py`` and ``scripts/recon/apgd_pycsou.py``.

3D data is provided in LenslessPiCam, but it is simulated. Real example data can be obtained from `Waller Lab <https://github.com/Waller-Lab/DiffuserCam/tree/master/example_data>`_.
For both the simulated data and the data from Waller Lab, it is best to set ``downsample=1``:

.. code:: bash

    python scripts/recon/gradient_descent.py \\
    input.psf="path/to/3D/psf.npy" \\
    input.data="path/to/lensless/data.tiff" \\
    preprocess.downsample=1


Other approaches
----------------

Scripts for other reconstruction algorithms can be found in ``scripts/recon`` and their
corresponding configurations in ``configs``.


**References**

.. [1] Boyd, S., Parikh, N., & Chu, E. (2011). Distributed optimization and statistical learning via the alternating direction method of multipliers. Now Publishers Inc.

"""


import abc
import numpy as np
import pathlib as plib
import matplotlib.pyplot as plt
from lensless.utils.plot import plot_image
from lensless.utils.io import get_dtype
from lensless.recon.rfft_convolve import RealFFTConvolve2D

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class ReconstructionAlgorithm(abc.ABC):
    """
    Abstract class for defining lensless imaging reconstruction algorithms.

    The following abstract methods need to be defined:

    * :py:class:`~lensless.ReconstructionAlgorithm._update`: updating state variables at each iterations.
    * :py:class:`~lensless.ReconstructionAlgorithm.reset`: reset state variables.
    * :py:class:`~lensless.ReconstructionAlgorithm._form_image`: any pre-processing that needs to be done in order to view the image estimate, e.g. reshaping or clipping.

    One advantage of deriving from this abstract class is that functionality for
    iterating, saving, and visualization is already implemented, namely in the
    ``apply`` method.

    Consequently, using a reconstruction algorithm that derives from it boils down
    to three steps:

    #. Creating an instance of the reconstruction algorithm.
    #. Setting the data.
    #. Applying the algorithm.


    """

    def __init__(
        self,
        psf,
        dtype=None,
        pad=True,
        n_iter=100,
        initial_est=None,
        reset=True,
        denoiser=None,
        **kwargs,
    ):
        """
        Base constructor. Derived constructor may define new state variables
        here and also reset them in `reset`.

        Parameters
        ----------

            psf : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
                Point spread function (PSF) that models forward propagation.
                Must be of shape (depth, height, width, channels) even if
                depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
                to load a PSF from a file such that it is in the correct format.
            dtype : float32 or float64
                Data type to use for optimization.
            pad : bool, optional
                Whether data needs to be padded prior to convolution. User may wish to
                optimize padded data and set this to False, as is done for :py:class:`~lensless.ADMM`.
                Defaults to True.
            n_iter : int, optional
                Number of iterations to run algorithm for. Can be overridden in `apply`.
            initial_est : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`, optional
                Initial estimate of the image. If not provided, the initial estimate is
                set to zero or to the mean of the data, depending on the algorithm.
            reset : bool, optional
                Whether to reset state variables in the base constructor. Defaults to True.
                If False, you should call reset() at one point to initialize state variables.
            denoiser : dict, optional
                Dictionary defining a denoiser for plug-and-play. Must contain the following keys:

                * ``"network"``: model to use as a denoiser.
                * ``"noise_level"``: noise level of the denoiser.

                If provided, the denoiser will be used as a projection function at each iteration.
                Defaults to None.
        """
        super().__init__()
        self.is_torch = False

        if torch_available:
            self.is_torch = isinstance(psf, torch.Tensor)

        assert len(psf.shape) == 4, "PSF must be 4D: (depth, height, width, channels)."
        assert psf.shape[3] == 3 or psf.shape[3] == 1, "PSF must either be rgb (3) or grayscale (1)"
        self._psf = psf
        self._npix = np.prod(self._psf.shape)
        self._n_iter = n_iter

        self._psf_shape = np.array(self._psf.shape)

        # set dtype
        dtype = get_dtype(dtype, self.is_torch)

        if self.is_torch:
            if dtype:
                self._psf = self._psf.type(dtype)
                self._dtype = dtype
            else:
                self._dtype = self._psf.dtype
            if self._dtype == torch.float32 or dtype == "float32":
                self._complex_dtype = torch.complex64
            elif self._dtype == torch.float64 or dtype == "float64":
                self._complex_dtype = torch.complex128
            else:
                raise ValueError(f"Unsupported dtype : {self._dtype}")

        else:
            if dtype:
                self._psf = self._psf.astype(dtype)
                self._dtype = dtype
            else:
                self._dtype = self._psf.dtype
            if self._dtype == np.float32 or dtype == "float32":
                self._complex_dtype = np.complex64
            elif self._dtype == np.float64 or dtype == "float64":
                self._complex_dtype = np.complex128
            else:
                raise ValueError(f"Unsupported dtype : {self._dtype}")

        self._convolver_param = {"dtype": dtype, "pad": pad, **kwargs}
        self._convolver = RealFFTConvolve2D(psf, dtype=dtype, pad=pad, **kwargs)
        self._padded_shape = self._convolver._padded_shape

        if pad:
            self._image_est_shape = self._psf_shape
        else:
            self._image_est_shape = self._convolver._padded_shape

        # pre-compute operators / outputs / set estimates
        if initial_est is not None:
            self._set_initial_estimate(initial_est)
        else:
            self._initial_est = None
        self._data = None

        self._denoiser = None
        if denoiser is not None:
            assert self.is_torch
            assert "network" in denoiser.keys()
            assert "noise_level" in denoiser.keys()

            from lensless.recon.utils import get_drunet_function_v2, load_drunet

            device = self._psf.device
            if denoiser["network"] == "DruNet":
                denoiser_model = load_drunet(requires_grad=False).to(device)
                self._denoiser = get_drunet_function_v2(denoiser_model, mode="inference")
            else:
                raise NotImplementedError(f"Unsupported denoiser: {denoiser['network']}")
            self._denoiser_noise_level = denoiser["noise_level"]

        # used inside trainable recon
        self.compensation_branch = None
        self.compensation_branch_inputs = None

        if reset:
            self.reset()

    @abc.abstractmethod
    def reset(self):
        """
        Reset state variables.
        """
        return

    @abc.abstractmethod
    def _update(self, iter):
        """
        Update state variables.
        """
        return

    @abc.abstractmethod
    def _form_image(self):
        """
        Any pre-processing to form a viewable image, e.g. reshaping or clipping.
        """
        return

    def set_data(self, data):
        """
        Set lensless data for recontruction.

        Parameters
        ----------
        data : :py:class:`~numpy.ndarray`
            Lensless data on which to iterate to recover an estimate of the
            scene. Should match provide PSF, i.e. shape and 2D (grayscale) or
            3D (RGB).
        """

        if self.is_torch:
            assert isinstance(data, torch.Tensor)
        else:
            assert isinstance(data, np.ndarray)

        assert len(data.shape) >= 3, "Data must be at least 3D: [..., width, height, channel]."

        # assert same shapes
        assert np.all(
            self._psf_shape[-3:-1] == np.array(data.shape)[-3:-1]
        ), "PSF and data shape mismatch"

        if len(data.shape) == 3:
            self._data = data[None, None, ...]
        elif len(data.shape) == 4:
            self._data = data[None, ...]
        else:
            self._data = data

    def _set_initial_estimate(self, image_est):
        """
        Set initial estimate of image, e.g. to warm-start algorithm.

        Note that reset() should be called after setting the initial estimate
        so that it is taken into account.

        Parameters
        ----------
        image_est : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Initial estimate of the image. Should match provide PSF shape.
        """

        if self.is_torch:
            assert isinstance(image_est, torch.Tensor)
        else:
            assert isinstance(image_est, np.ndarray)

        assert (
            len(image_est.shape) >= 4
        ), "Image estimate must be at least 4D: [..., depth, width, height, channel]."

        # assert same shapes
        assert np.all(
            self._image_est_shape[-3:-1] == np.array(image_est.shape)[-3:-1]
        ), f"Image estimate must be of shape (..., width, height, channel): {self._image_est_shape[-3:-1]}"

        if len(image_est.shape) == 4:
            self._initial_est = image_est[None, ...]
        else:
            self._initial_est = image_est

    def set_image_estimate(self, image_est):
        """
        Set initial estimate of image, e.g. to warm-start algorithm.

        Parameters
        ----------
        image_est : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Initial estimate of the image. Should match provide PSF shape.
        """

        if self.is_torch:
            assert isinstance(image_est, torch.Tensor)
        else:
            assert isinstance(image_est, np.ndarray)

        assert (
            len(image_est.shape) >= 4
        ), "Image estimate must be at least 4D: [..., depth, width, height, channel]."

        # assert same shapes
        assert np.all(
            self._image_est_shape[-3:-1] == np.array(image_est.shape)[-3:-1]
        ), f"Image estimate must be of shape (..., width, height, channel): {self._image_est_shape[-3:-1]}"

        if len(image_est.shape) == 4:
            self._image_est = image_est[None, ...]
        else:
            self._image_est = image_est

    def get_image_estimate(self):
        """Get current image estimate as [Batch, Depth, Height, Width, Channels]."""
        return self._form_image()

    def _set_psf(self, psf):
        """
        Set PSF.

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            PSF to set.
        """
        assert (
            psf.shape[-1] == 3 or psf.shape[-1] == 1
        ), "PSF must either be rgb (3) or grayscale (1)"
        assert self._psf.shape == psf.shape, "new PSF must have same shape as old PSF"
        assert isinstance(psf, type(self._psf)), "new PSF must have same type as old PSF"

        self._psf = psf
        self._convolver = RealFFTConvolve2D(
            psf,
            dtype=self._convolver._psf.dtype,
            pad=self._convolver.pad,
            norm=self._convolver.norm,
        )
        self.reset()

    def _progress(self):
        """
        Optional method for printing progress update, e.g. relative improvement
        in reconstruction.
        """
        return

    def _get_numpy_data(self, data):
        """
        Extract data from torch or numpy array.

        Parameters
        ----------
        data : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Data to extract.

        Returns
        -------
        :py:class:`~numpy.ndarray`
            Extracted data.
        """
        if self.is_torch:
            return data.detach().cpu().numpy()
        else:
            return data

    def apply(
        self,
        n_iter=None,
        disp_iter=-1,
        plot_pause=0.2,
        plot=False,
        save=False,
        gamma=None,
        ax=None,
        reset=True,
        background=None,
        **kwargs,
    ):
        """
        Method for performing iterative reconstruction. Note that `set_data`
        must be called beforehand.

        Parameters
        ----------
        n_iter : int, optional
            Number of iterations. If not provided, default to `self._n_iter`.
        disp_iter : int
            How often to display and/or intermediate reconstruction (in number
            of iterations). If `None` OR `plot` or `save` are False, no
            intermediate reconstruction will be plotted/saved.
        plot_pause : float
            Number of seconds to pause after displaying reconstruction.
        plot : bool
            Whether to plot final result, and intermediate results if
            `disp_iter` is not None.
        save : bool
            Whether to save final result (as PNG), and intermediate results if
            `disp_iter` is not None.
        gamma : float, optional
            Gamma correction factor to apply for plots. Default is None.
        ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill for plotting/saving, default is to create one.
        reset : bool, optional
            Whether to reset state variables before applying reconstruction. Default to True.
            Set to false if continuing reconstruction from previous state.

        Returns
        -------
        final_im : :py:class:`~numpy.ndarray`
            Final reconstruction.
        ax : :py:class:`~matplotlib.axes.Axes`
            `Axes` object on which final reconstruction is displayed. Only
            returning if `plot` or `save` is True.

        """
        assert self._data is not None, "Must set data with `set_data()`"
        assert (
            self._data.shape[0] == 1
        ), "Apply doesn't supports processing multiple images at once."

        if background is not None:
            self._data = self._data - background
            self._data[self._data < 0] = 0

        if reset:
            self.reset()

        if n_iter is None:
            n_iter = self._n_iter

        if (plot or save) and disp_iter is not None:
            if ax is None:
                img = self._form_image()
                ax = plot_image(self._get_numpy_data(img[0]), gamma=gamma)

        else:
            ax = None
            disp_iter = n_iter + 1

        if self.compensation_branch is not None:
            self.compensation_branch_inputs = [self._data]

        for i in range(n_iter):
            self._update(i)
            if self.compensation_branch is not None and i < self._n_iter - 1:
                self.compensation_branch_inputs.append(self._form_image())

            if (plot or save) and (i + 1) % disp_iter == 0:
                self._progress()
                img = self._form_image()
                ax = plot_image(self._get_numpy_data(img[0]), ax=ax, gamma=gamma)
                if hasattr(ax, "__len__"):
                    ax[0, 0].set_title("Reconstruction after iteration {}".format(i + 1))
                else:
                    ax.set_title("Reconstruction after iteration {}".format(i + 1))
                if save:
                    plt.savefig(plib.Path(save) / f"{i + 1}.png")
                if plot:
                    plt.draw()
                    plt.pause(plot_pause)

        final_im = self._form_image()[0]
        if plot:
            ax = plot_image(self._get_numpy_data(final_im), ax=ax, gamma=gamma)
            if hasattr(ax, "__len__"):
                ax[0, 0].set_title("Final reconstruction after {} iterations".format(n_iter))
            else:
                ax.set_title("Final reconstruction after {} iterations".format(n_iter))
            if save:
                plt.savefig(plib.Path(save) / f"{n_iter}.png")
            return final_im, ax
        else:
            return final_im

    def reconstruction_error(self, prediction=None, lensless=None, psfs=None, normalize=True):
        """
        Compute reconstruction error.

        Parameters
        ----------
        prediction :  :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`, optional
            Reconstructed image. If None, use current estimate, default None.
        lensless : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`, optional
            Lensless image. If None, use data provided by `set_data()`, default None.

        Returns
        -------
        _type_
            _description_
        """
        # default to current estimate and data if not provided
        if prediction is None:
            prediction = self.get_image_estimate()
        if lensless is None:
            lensless = self._data

        # convolver = self._convolver
        if psfs is None:
            psfs = self._psf
        convolver = RealFFTConvolve2D(psfs.to(prediction.device), **self._convolver_param)
        if not convolver.pad:
            prediction = convolver._pad(prediction)
        Hx = convolver.convolve(prediction)

        if not convolver.pad:
            Hx = convolver._crop(Hx)

        # -- normalize
        if normalize:
            min_vals = torch.amin(Hx, dim=(-1, -2, -3), keepdim=True)
            Hx = Hx - min_vals
            max_vals = torch.amax(Hx, dim=(-1, -2, -3), keepdim=True)
            Hx = Hx / max_vals

        # don't reduce batch dimension
        if self.is_torch:
            # torch.mean((Hx - lensless) ** 2, dim=(-1, -2, -3, -4))
            return torch.sum((Hx - lensless) ** 2, dim=(-1, -2, -3, -4)) / self._npix

        else:
            return np.sum((Hx - lensless) ** 2, axis=(-1, -2, -3, -4)) / self._npix



"""
Apply gradient descent.

```
python scripts/recon/gradient_descent.py
```

"""

import hydra
from hydra.utils import to_absolute_path
import os
import numpy as np
import time
import pathlib as plib
import matplotlib.pyplot as plt
from lensless.utils.io import load_data
from lensless import (
    GradientDescentUpdate,
    GradientDescent,
    NesterovGradientDescent,
    FISTA,
)


@hydra.main(version_base=None, config_path="../../configs/recon", config_name="defaults")
def gradient_descent(
    config,
):
    psf, data = load_data(
        psf_fp=to_absolute_path(config.input.psf),
        data_fp=to_absolute_path(config.input.data),
        dtype=config.input.dtype,
        downsample=config["preprocess"]["downsample"],
        bayer=config["preprocess"]["bayer"],
        blue_gain=config["preprocess"]["blue_gain"],
        red_gain=config["preprocess"]["red_gain"],
        plot=config["display"]["plot"],
        flip=config["preprocess"]["flip"],
        gamma=config["display"]["gamma"],
        gray=config["preprocess"]["gray"],
        single_psf=config["preprocess"]["single_psf"],
        shape=config["preprocess"]["shape"],
        use_torch=config.torch,
        torch_device=config.torch_device,
    )

    disp = config["display"]["disp"]
    if disp < 0:
        disp = None

    save = config["save"]
    if save:
        save = os.getcwd()

    start_time = time.time()

    if config["gradient_descent"]["method"] == GradientDescentUpdate.VANILLA:
        recon = GradientDescent(psf)
    elif config["gradient_descent"]["method"] == GradientDescentUpdate.NESTEROV:
        recon = NesterovGradientDescent(
            psf,
            p=config["gradient_descent"]["nesterov"]["p"],
            mu=config["gradient_descent"]["nesterov"]["mu"],
        )
    else:
        recon = FISTA(
            psf,
            tk=config["gradient_descent"]["fista"]["tk"],
        )

    recon.set_data(data)
    print(f"Setup time : {time.time() - start_time} s")

    start_time = time.time()
    res = recon.apply(
        n_iter=config["gradient_descent"]["n_iter"],
        disp_iter=disp,
        save=save,
        gamma=config["display"]["gamma"],
        plot=config["display"]["plot"],
    )
    print(f"Processing time : {time.time() - start_time} s")

    if config.torch:
        img = res[0].cpu().numpy()
    else:
        img = res[0]

    if config["display"]["plot"]:
        plt.show()
    if save:
        np.save(plib.Path(save) / "final_reconstruction.npy", img)
        print(f"Files saved to : {save}")


if __name__ == "__main__":
    gradient_descent()



"""

Simulate a mask, simulate a few measurements with it, and reconstruct the images.

Procedure is as follows:

1) Simulate the mask.
2) Simulate measurements with the mask and specified physical parameters.
3) Reconstruct the images from the measurements.

Example usage:

Simulate FlatCam with separable simulation and Tikhonov reconstuction (https://arxiv.org/abs/1509.00116, Eq 7):
```
python scripts/sim/mask_dataset.py mask.type=MLS simulation.flatcam=True recon.algo=tikhonov
```

Simulate FlatCam with PSF simulation and Tikhonov reconstuction:
```
python scripts/sim/mask_dataset.py mask.type=MLS simulation.flatcam=False recon.algo=tikhonov
```

Simulate FlatCam with PSF simulation and ADMM reconstruction:
```
python scripts/sim/mask_dataset.py mask.type=MLS simulation.flatcam=False recon.algo=admm
```

Simulate Fresnel Zone Aperture camera with PSF simulation and ADMM reconstuction (https://www.nature.com/articles/s41377-020-0289-9):
```
python scripts/sim/mask_dataset.py mask.type=FZA recon.algo=admm
```

Simulate PhaseContour camera with PSF simulation and ADMM reconstuction (https://ieeexplore.ieee.org/document/9076617):
```
python scripts/sim/mask_dataset.py mask.type=PhaseContour recon.algo=admm
```

If Pytorch is installed, you can use the `torch` flag to use Pytorch for the reconstruction (ADMM only):
```
python scripts/sim/mask_dataset.py mask.type=PhaseContour recon.algo=admm
```

"""

import hydra
import warnings
from hydra.utils import to_absolute_path
from lensless.utils.io import load_image, save_image
from lensless.utils.image import rgb2gray
import numpy as np
from lensless import ADMM
from lensless.eval.metric import mse, psnr, ssim, lpips
from waveprop.simulation import FarFieldSimulator
import glob
import os
from tqdm import tqdm
from lensless.hardware.mask import CodedAperture, PhaseContour, FresnelZoneAperture
from lensless.recon.tikhonov import CodedApertureReconstruction


@hydra.main(version_base=None, config_path="../../configs/simulate", config_name="mask_sim_dataset")
def simulate(config):

    if config.torch:
        try:
            import torch
        except ImportError:
            raise ImportError("Pytorch not found. Please install pytorch to use torch mode.")

    # set seed
    np.random.seed(config.seed)

    mask2sensor = config.simulation.mask2sensor
    sensor = config.simulation.sensor
    snr_db = config.simulation.snr_db
    downsample = config.simulation.downsample

    dataset = to_absolute_path(config.files.dataset)
    if not os.path.isdir(dataset):
        print(f"No dataset found at {dataset}")
        try:
            from torchvision.datasets.utils import download_and_extract_archive, download_url
        except ImportError:
            exit()
        msg = "Do you want to download the sample CelebA dataset (764KB)?"

        # default to yes if no input is given
        valid = input("%s (Y/n) " % msg).lower() != "n"
        if valid:
            url = "https://drive.switch.ch/index.php/s/Q5OdDQMwhucIlt8/download"
            filename = "celeb_mini.zip"
            download_and_extract_archive(
                url, os.path.dirname(dataset), filename=filename, remove_finished=True
            )

    mask_type = config.mask.type

    # check for flatcam simulation
    flatcam_sim = config.simulation.flatcam
    if flatcam_sim and mask_type.upper() not in ["MURA", "MLS"]:
        warnings.warn(
            "Flatcam simulation only supported for MURA and MLS masks. Using far field simulation with PSF."
        )
        flatcam_sim = False

    if config.save:
        if flatcam_sim:
            save_dir = to_absolute_path(config.files.dataset + "_" + mask_type + "_flatcam_sim")
        else:
            save_dir = to_absolute_path(config.files.dataset + "_" + mask_type)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
            os.makedirs(os.path.join(save_dir, "sensor_plane"))
            os.makedirs(os.path.join(save_dir, "object_plane"))
            os.makedirs(os.path.join(save_dir, "reconstruction"))

    # simulate mask
    mask = None
    if mask_type.upper() in ["MURA", "MLS"]:
        mask = CodedAperture.from_sensor(
            sensor_name=sensor,
            downsample=downsample,
            method=mask_type,
            distance_sensor=mask2sensor,
            **config.mask,
        )
        psf_sim = mask.psf / np.linalg.norm(mask.psf.ravel())
    elif mask_type.upper() == "FZA":
        mask = FresnelZoneAperture.from_sensor(
            sensor_name=sensor,
            downsample=downsample,
            distance_sensor=mask2sensor,
            **config.mask,
        )
        psf_sim = mask.psf / np.linalg.norm(mask.psf.ravel())
    elif mask_type == "PhaseContour":
        mask = PhaseContour.from_sensor(
            sensor_name=sensor,
            downsample=downsample,
            distance_sensor=mask2sensor,
            n_iter=config.mask.phase_mask_iter,
            **config.mask,
        )
        psf_sim = mask.psf / np.linalg.norm(mask.psf.ravel())
    assert mask is not None, f"Mask type {mask_type} not implemented."

    if config.simulation.grayscale and len(psf_sim.shape) == 3:
        psf_sim = rgb2gray(psf_sim)

    if config.simulation.downsample > 1:
        print(f"Downsampled to {psf_sim.shape}.")

    # prepare simulator object
    simulator = FarFieldSimulator(psf=psf_sim, **config.simulation)

    # loop over files in dataset
    print("\nSimulating dataset...")
    files = glob.glob(os.path.join(dataset, f"*.{config.files.image_ext}"))
    if config.files.n_files is not None:
        files = files[: config.files.n_files]

    for fp in tqdm(files):

        # load image as numpy array
        image = load_image(fp) / 255
        if config.simulation.grayscale and len(image.shape) == 3:
            image = rgb2gray(image)

        # simulate image
        image_plane, object_plane = simulator.propagate(image, return_object_plane=True)
        if flatcam_sim:
            image_plane = mask.simulate(object_plane, snr_db=snr_db)

        if config.save:

            bn = os.path.basename(fp).split(".")[0] + ".png"

            # can serve as ground truth
            object_plane_fp = os.path.join(save_dir, "object_plane", bn)
            save_image(object_plane, object_plane_fp)  # use max range of 255

            # lensless image
            lensless_fp = os.path.join(save_dir, "sensor_plane", bn)
            save_image(image_plane, lensless_fp, max_val=config.simulation.max_val)

    # reconstruction
    recon_algo = config.recon.algo.lower()

    if config.recon.algo is not None:

        print("\nReconstructing lensless measurements...")
        # -- initialize reconstruction object
        if recon_algo == "tikhonov":
            if config.torch:
                raise NotImplementedError("Tikhonov reconstruction not implemented for torch.")
            recon = CodedApertureReconstruction(
                mask, object_plane.shape, lmbd=config.recon.tikhonov.reg
            )
        elif recon_algo == "admm":
            psf = psf_sim[np.newaxis, :, :, :]
            if config.torch:
                psf = torch.from_numpy(psf).to(config.torch_device)
            recon = ADMM(psf, **config.recon.admm)

        # -- metrics
        mse_vals = []
        psnr_vals = []
        ssim_vals = []
        if not config.simulation.grayscale:
            lpips_vals = []
        else:
            lpips_vals = None

        # -- loop over files in dataset
        files = glob.glob(os.path.join(save_dir, "sensor_plane", "*.png"))
        if config.files.n_files is not None:
            files = files[: config.files.n_files]

        for fp in tqdm(files):

            lensless = load_image(fp, as_4d=True)
            lensless = lensless / np.max(lensless)
            if recon_algo == "tikhonov":
                recovered = recon.apply(lensless[0])
            elif recon_algo == "admm":
                if config.torch:
                    lensless = torch.from_numpy(lensless).to(config.torch_device)
                recon.set_data(lensless)
                res, _ = recon.apply(n_iter=config.recon.admm.n_iter)

                # get first depth
                if config.torch:
                    recovered = res[0].cpu().numpy()
                else:
                    recovered = res[0]

            if config.save:
                bn = os.path.basename(fp).split(".")[0] + ".png"
                lensless_fp = os.path.join(save_dir, "reconstruction", bn)

                save_image(recovered, lensless_fp, max_val=config.simulation.max_val)

            # compute metrics
            object_plane_fp = os.path.join(save_dir, "object_plane", os.path.basename(fp))
            object_plane = load_image(object_plane_fp)

            mse_vals.append(mse(object_plane, recovered))
            psnr_vals.append(psnr(object_plane, recovered))
            if config.simulation.grayscale:
                ssim_vals.append(ssim(object_plane, recovered, channel_axis=None))
            else:
                ssim_vals.append(ssim(object_plane, recovered))
            if lpips_vals is not None:
                lpips_vals.append(lpips(object_plane, recovered))

    print("\nMSE (avg)", np.mean(mse_vals))
    print("PSNR (avg)", np.mean(psnr_vals))
    print("SSIM (avg)", np.mean(ssim_vals))
    if lpips_vals is not None:
        print("LPIPS (avg)", np.mean(lpips_vals))

    print("Results saved to", save_dir)


if __name__ == "__main__":
    simulate()



import hydra
import yaml
import torch
from lensless import ADMM
from lensless.utils.dataset import HFDataset
import os
from lensless.utils.io import save_image
from tqdm import tqdm
from lensless.recon.model_dict import download_model, load_model
import numpy as np
from torchmetrics import StructuralSimilarityIndexMeasure
from torchmetrics.image import lpip, psnr
import json
from matplotlib import pyplot as plt


def key_to_ratio_correct(key_length, bit_depth, n_pixel):
    return np.emath.logn(bit_depth, 2) * key_length / n_pixel


@hydra.main(
    version_base=None, config_path="../../configs/recon", config_name="digicam_mirflickr_psf_err"
)
def apply_pretrained(config):
    device = config.device
    model_name = config.model
    percent_pixels_wrong = config.percent_pixels_wrong
    key_to_ratio = {"DigiCam": config.digicam_ratio}

    if config.metrics_fp is not None:

        # load metrics from file
        with open(config.metrics_fp, "r") as f:
            metrics_values = json.load(f)

        # # if not normalized... all PSFs have roughtly same norm
        # metrics_values["psf_err"] = np.array(metrics_values["psf_err"]) / 1.7302357e-06
        # metrics_values["psf_err"] = metrics_values["psf_err"].tolist()

        # # resave metrics dict to JSON
        # with open(config.metrics_fp, "w") as f:
        #     json.dump(metrics_values, f, indent=4)

    else:

        # load config
        if model_name == "admm":
            # take config from unrolled ADMM for dataset
            model_path = download_model(
                camera="digicam", dataset="mirflickr_multi_25k", model="Unet4M+U5+Unet4M_wave"
            )
            config_path = os.path.join(model_path, ".hydra", "config.yaml")
            with open(config_path, "r") as stream:
                model_config = yaml.safe_load(stream)

        else:
            model_path = download_model(
                camera="digicam", dataset="mirflickr_multi_25k", model=model_name
            )
            config_path = os.path.join(model_path, ".hydra", "config.yaml")
            with open(config_path, "r") as stream:
                model_config = yaml.safe_load(stream)

        metrics = {
            "PSNR": psnr.PeakSignalNoiseRatio(reduction=None, dim=(1, 2, 3), data_range=(0, 1)).to(
                device
            ),
            "SSIM": StructuralSimilarityIndexMeasure(reduction=None, data_range=(0, 1)).to(device),
            "LPIPS_Vgg": lpip.LearnedPerceptualImagePatchSimilarity(
                net_type="vgg", normalize=True, reduction="sum"
            ).to(device),
            "psf_err": torch.nn.functional.mse_loss,
        }

        # load data
        # TODO missing simulation parameters???
        test_set = HFDataset(
            huggingface_repo=(
                model_config["files"]["dataset"] if config.hf_repo is None else config.hf_repo
            ),
            psf=(
                model_config["files"]["huggingface_psf"]
                if "huggingface_psf" in model_config["files"]
                else None
            ),
            split="test",
            display_res=model_config["files"]["image_res"],
            rotate=model_config["files"]["rotate"],
            flipud=model_config["files"]["flipud"],
            flip_lensed=model_config["files"]["flip_lensed"],
            downsample=model_config["files"]["downsample"],
            alignment=model_config["alignment"],
            simulation_config=model_config["simulation"],
            force_rgb=model_config["files"].get("force_rgb", False),
            cache_dir=config.cache_dir,
            save_psf=False,
            return_mask_label=True,
        )

        psf_norms = []
        for mask_label in test_set.psf.keys():
            psf_norms.append(np.mean(test_set.psf[mask_label].cpu().numpy().flatten() ** 2))
        psf_norms = np.array(psf_norms)

        n_files = config.n_files
        if n_files is None:
            n_files = len(test_set)
        percent_pixels_wrong = config.percent_pixels_wrong

        # initialize metrics dict
        metrics_values = {k: np.zeros((len(percent_pixels_wrong), n_files)) for k in metrics.keys()}

        for i in config.save_idx:
            # make folder
            save_dir = str(i)
            os.makedirs(save_dir, exist_ok=True)

        for idx in tqdm(range(n_files)):

            # get data
            lensless, lensed, mask_label = test_set[idx]
            lensless = lensless.to(device)

            if idx in config.save_idx:
                if lensed is not None:
                    lensed_np = lensed[0].cpu().numpy()
                    save_image(lensed_np, os.path.join(str(idx), f"original_idx{idx}.png"))
                save_image(
                    lensless[0].cpu().numpy(), os.path.join(str(idx), f"lensless_idx{idx}.png")
                )

            # -- reshape for torchmetrics
            lensed = lensed.reshape(-1, *lensed.shape[-3:]).movedim(-1, -3)
            lensed_max = torch.amax(lensed, dim=(1, 2, 3), keepdim=True)
            lensed = lensed / lensed_max
            lensed = lensed.to(device)

            _metrics_idx = {k: [] for k in metrics.keys()}

            for percent_wrong in percent_pixels_wrong:

                # perturb mask
                mask_vals = test_set.get_mask_vals(mask_label)

                noisy_mask_vals = mask_vals.copy()
                if percent_wrong > 0:

                    n_pixels = mask_vals.size
                    assert n_pixels == config.n_pixel

                    n_wrong_pixels = int(n_pixels * percent_wrong / 100)
                    wrong_pixels = np.random.choice(n_pixels, n_wrong_pixels, replace=False)
                    noisy_mask_vals = noisy_mask_vals.flatten()

                    if config.flip:
                        noisy_mask_vals[wrong_pixels] = (
                            1 - noisy_mask_vals[wrong_pixels]
                        )  # flip pixel value
                    else:
                        # reset values randomly
                        noisy_mask_vals[wrong_pixels] = np.random.uniform(size=n_wrong_pixels)
                    noisy_mask_vals = noisy_mask_vals.reshape(mask_vals.shape)

                    # noise = np.random.uniform(size=mask_vals.shape)
                    # # -- rescale noise to desired SNR
                    # mask_var = ndimage.variance(mask_vals)
                    # noise_var = ndimage.variance(noise)
                    # fact = np.sqrt(mask_var / noise_var / (10 ** (mask_snr_db / 10)))
                    # noisy_mask_vals = mask_vals + fact * noise
                    # # -- clip to [0, 1]
                    # noisy_mask_vals = np.clip(noisy_mask_vals, 0, 1)

                # simulate PSF
                psf = test_set.simulate_psf(noisy_mask_vals)
                psf = psf.to(device)

                # compute L2 error with normal PSF
                _metrics_idx["psf_err"].append(
                    metrics["psf_err"](psf, test_set.psf[mask_label].to(device)).item()
                    / psf_norms[mask_label]
                )

                # load model
                if model_name == "admm":
                    recon = ADMM(psf, n_iter=config.n_iter)
                else:
                    # load best model
                    recon = load_model(model_path, psf, device, verbose=False)

                # reconstruct
                with torch.no_grad():
                    recon.set_data(lensless)
                    res = recon.apply(
                        disp_iter=-1,
                        save=False,
                        gamma=None,
                        plot=False,
                    )
                recon = res[0]

                # prepare for metrics
                # -- convert to [N*D, C, H, W] for torchmetrics
                prediction = recon.reshape(-1, *recon.shape[-3:]).movedim(-1, -3)
                # - extract ROI
                prediction = test_set.extract_roi(prediction, axis=(-2, -1))
                # -- normalize
                prediction_max = torch.amax(prediction, dim=(1, 2, 3), keepdim=True)
                prediction = prediction / prediction_max

                for k, metric in metrics.items():
                    if k == "psf_err":
                        continue
                    _metrics_idx[k].append(metric(prediction, lensed).item())

                # save
                if idx in config.save_idx:

                    # PSF
                    from lensless.utils.image import gamma_correction

                    psf_np = psf.cpu().numpy().squeeze()
                    fp = os.path.join(str(idx), f"psf_percentwrong{percent_wrong}.png")
                    psf_np /= np.max(psf_np)
                    psf_np = gamma_correction(psf_np, gamma=3)
                    save_image(psf_np, fp)

                    # reconstruction
                    img = recon.cpu().numpy().squeeze()
                    alignment = test_set.alignment
                    top_left = alignment["top_left"]
                    height = alignment["height"]
                    width = alignment["width"]
                    res_np = img[
                        top_left[0] : top_left[0] + height, top_left[1] : top_left[1] + width
                    ]
                    fp = os.path.join(str(idx), f"{model_name}_percentwrong{percent_wrong}.png")
                    save_image(res_np, fp)

            # save metrics
            for k, v in _metrics_idx.items():
                metrics_values[k][:, idx] = v

        # save metric dict to JSON
        # -- make sure to convert numpy arrays to lists
        for k, v in metrics_values.items():
            metrics_values[k] = v.tolist()
        with open(f"{model_name}_metrics.json", "w") as f:
            json.dump(metrics_values, f, indent=4)

    # plot each metrics
    ## - config text size
    plt.rcParams.update({"font.size": 24})
    ## - config line width
    plt.rcParams.update({"lines.linewidth": 5})

    key_lengths = config.compare_aes
    for key_length in key_lengths:
        key_to_ratio[key_length] = key_to_ratio_correct(
            key_length, config.bit_depth, config.n_pixel
        )
        if config.plot_vs_percent_wrong:
            key_to_ratio[key_length] = 1 - key_to_ratio[key_length]
    linestyles = ["--", "-.", ":"]
    colors = ["red", "forestgreen", "purple", "green", "blue", "black"]

    for k, v in metrics_values.items():
        plt.figure(figsize=(6.7, 5))
        y_vals = np.mean(v, axis=1)
        if config.plot_vs_percent_wrong:
            plt.xlabel("Percent pixels wrong [%]")
            x_vals = percent_pixels_wrong
        else:
            plt.xlabel("Percent pixels correct [%]")
            x_vals = 100 - np.array(percent_pixels_wrong)
        plt.xlim([0, 100])
        if k == "psf_err":
            plt.plot(x_vals, y_vals * 100)
            plt.ylabel("Relative PSF error [%]")
            plt.ylim([0, 100])
        elif k == "PSNR":
            plt.plot(x_vals, y_vals)
            plt.ylabel("PSNR [dB]")
        elif k == "LPIPS_Vgg":
            plt.plot(x_vals, y_vals)
            plt.ylabel("LPIPS")
        else:
            plt.plot(x_vals, y_vals)
            plt.ylabel(k)

        if config.plot_vs_percent_wrong:
            print(f"-- Metric {k} : ", y_vals[::-1])
            print("% wrong vals : ", x_vals[::-1])
        else:
            print(f"-- Metric {k} : ", y_vals[::-1])
            print("% corrects vals : ", x_vals[::-1])

        # plot keys
        # for idx, key_length in enumerate(key_lengths):
        for idx, _key in enumerate(list(key_to_ratio.keys())):
            if isinstance(_key, int):
                label = f"AES-{_key}"
            else:
                label = _key
            plt.axvline(
                key_to_ratio[_key] * 100, color=colors[idx], linestyle=linestyles[idx], label=label
            )
            # label along x axis
            # plt.text(
            #     key_to_ratio[key_length] * 100,
            #     plt.ylim()[-1],
            #     # f"{key_length} bits ({100 * key_to_ratio[key_length]:.2f}%)",
            #     f"{key_length} bits",
            #     rotation=45,
            #     verticalalignment="bottom",
            # )

        # legend bottom right
        if k == "PSNR" or k == "SSIM":
            plt.legend(loc="lower right")
        else:
            plt.legend(loc="upper right")

        # save plot
        # - tight
        plt.grid()
        plt.tight_layout()
        plt.savefig(f"{k}_{model_name}.png", bbox_inches="tight")


if __name__ == "__main__":
    apply_pretrained()



# #############################################################################
# image.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Julien SAHLI [julien.sahli@epfl.ch]
# #############################################################################


import cv2
import scipy.signal
import numpy as np
from lensless.hardware.constants import RPI_HQ_CAMERA_CCM_MATRIX, RPI_HQ_CAMERA_BLACK_LEVEL

try:
    import torch
    import torchvision.transforms as tf
    from torchvision.transforms.functional import rgb_to_grayscale, rotate

    torch_available = True
except ImportError:
    torch_available = False

SUPPORTED_BIT_DEPTH = np.array([8, 10, 12, 16])
FLOAT_DTYPES = [np.float32, np.float64]


def resize(img, factor=None, shape=None, interpolation=cv2.INTER_CUBIC):
    """
    Resize by given factor.

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray`
        Image to downsample
    factor : int or float
        Resizing factor.
    shape : tuple
        Shape to copy ([depth,] height, width, color). If provided, (height, width) is used.
    interpolation : OpenCV interpolation method
        See https://docs.opencv.org/2.4/modules/imgproc/doc/geometric_transformations.html#cv2.resize

    Returns
    -------
    img : :py:class:`~numpy.ndarray`
        Resized image.
    """
    min_val = img.min()
    max_val = img.max()
    img_shape = np.array(img.shape)[-3:-1]

    assert not ((factor is None) and (shape is None)), "Must specify either factor or shape"
    new_shape = tuple(img_shape * factor) if shape is None else shape[-3:-1]
    new_shape = [int(i) for i in new_shape]

    if np.array_equal(img_shape, new_shape):
        return img

    if torch_available:
        # torch resize expects an input of form [color, depth, width, height]
        tmp = np.moveaxis(img, -1, 0)
        tmp = torch.from_numpy(tmp.copy())
        resized = tf.Resize(size=new_shape, antialias=True)(tmp).numpy()
        resized = np.moveaxis(resized, 0, -1)

    else:
        resized = np.array(
            [
                cv2.resize(img[i], dsize=tuple(new_shape[::-1]), interpolation=interpolation)
                for i in range(img.shape[-4])
            ]
        )
        # OpenCV discards channel dimension if it is 1, put it back
        if len(resized.shape) == 3:
            # resized = resized[:, :, :, np.newaxis]
            resized = np.expand_dims(resized, axis=-1)

    return np.clip(resized, min_val, max_val)


def rotate_HWC(img, angle):

    # to CHW
    img = img.movedim(-1, -3)

    remove_depth_dim = False
    if len(img.shape) == 5:
        # remove singleton depth dimension before, otherwise doesn't work..
        if img.shape[1] == 1:
            img = img.squeeze(1)
            remove_depth_dim = True

    # rotate
    img_rot = rotate(img, angle, expand=False)

    if remove_depth_dim:
        img_rot = img_rot.unsqueeze(1)

    # to HWC
    img_rot = img_rot.movedim(-3, -1)
    return img_rot


def shift_with_pad(img, shift, pad_mode="constant", axis=(0, 1)):
    n_dim = len(img.shape)

    # padding for shifting
    pad_width = [(0, 0) for _ in range(n_dim)]
    for i, s in zip(axis, shift):
        if s < 0:
            pad_width[i] = (0, -s)
        else:
            pad_width[i] = (s, 0)
    pad_width = tuple(pad_width)

    # slice after padding
    slice_obj = [slice(None) for _ in range(n_dim)]
    for i, s in zip(axis, shift):
        if s < 0:
            slice_obj[i] = slice(-s, None)
        elif s == 0:
            slice_obj[i] = slice(None)
        else:
            slice_obj[i] = slice(None, -s)

    if torch_available and isinstance(img, torch.Tensor):

        # concatenate all tuples into single one for torch padding
        # -- in reverse order as pytorch pads dimensions in reverse order
        pad_width_torch = tuple([item for sublist in pad_width[::-1] for item in sublist])
        shifted = torch.nn.functional.pad(img, pad=pad_width_torch, mode=pad_mode)

    else:
        shifted = np.pad(img, pad_width=pad_width, mode=pad_mode)

    shifted = shifted[tuple(slice_obj)]

    return shifted


def is_grayscale(img):
    """
    Check if image is RGB. Assuming image is of shape ([depth,] height, width, color).

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
        Image array.

    Returns
    -------
    bool
        Whether image is RGB.
    """
    return img.shape[-1] == 1


def rgb2gray(rgb, weights=None, keepchanneldim=True):
    """
    Convert RGB array to grayscale.

    Parameters
    ----------
    rgb : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
        ([Depth,] Height, Width, Channel) image.
    weights : :py:class:`~numpy.ndarray`
        [Optional] (3,) weights to convert from RGB to grayscale. Only used for NumPy arrays.
    keepchanneldim : bool
        Whether to keep the channel dimension. Default is True.

    Returns
    -------
    img :py:class:`~numpy.ndarray`
        Grayscale image of dimension ([depth,] height, width [, 1]).

    """

    use_torch = False
    if torch_available:
        if torch.is_tensor(rgb):
            use_torch = True

    if use_torch:

        # move channel dimension to third to last
        if len(rgb.shape) == 4:
            rgb = rgb.permute(0, 3, 1, 2)
        elif len(rgb.shape) == 3:
            rgb = rgb.permute(2, 0, 1)
        else:
            raise ValueError("Input must be at least 3D.")

        image = rgb_to_grayscale(rgb)

        # move channel dimension to last
        if len(rgb.shape) == 4:
            image = image.permute(0, 2, 3, 1)
        elif len(rgb.shape) == 3:
            image = image.permute(1, 2, 0)

        if not keepchanneldim:
            image = image.squeeze(-1)

        return image

    else:

        if weights is None:
            weights = np.array([0.299, 0.587, 0.114])
        assert len(weights) == 3

        if len(rgb.shape) == 4:
            image = np.tensordot(rgb, weights, axes=((3,), 0))
        elif len(rgb.shape) == 3:
            image = np.tensordot(rgb, weights, axes=((2,), 0))
        else:
            raise ValueError("Input must be at least 3D.")

        if keepchanneldim:
            return image[..., np.newaxis]
        else:
            return image


def gamma_correction(vals, gamma=2.2):
    """
    Apply `gamma correction <https://www.cambridgeincolour.com/tutorials/gamma-correction.htm>`__.

    Parameters
    ----------
    vals : :py:class:`~numpy.ndarray`
        RGB values to gamma correct.
    gamma : float, optional
        Gamma correction factor.

    Returns
    -------
    vals : :py:class:`~numpy.ndarray`
        Gamma-corrected data.

    """

    # Rec. 709 gamma correction
    # http://www.fourmilab.ch/documents/specrend/specrend.c
    cc = 0.018
    inv_gam = 1 / gamma
    clip_val = (1.099 * np.power(cc, inv_gam) - 0.099) / cc
    return np.where(vals < cc, vals * clip_val, 1.099 * np.power(vals, inv_gam) - 0.099)


def get_max_val(img, nbits=None):
    """
    For uint image.

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray`
        Image array.
    nbits : int, optional
        Number of bits per pixel. Detect if not provided.

    Returns
    -------
    max_val : int
        Maximum pixel value.
    """
    assert img.dtype not in FLOAT_DTYPES
    if nbits is None:
        nbits = int(np.ceil(np.log2(img.max())))

    if nbits not in SUPPORTED_BIT_DEPTH:
        nbits = SUPPORTED_BIT_DEPTH[nbits < SUPPORTED_BIT_DEPTH][0]
    max_val = 2**nbits - 1
    if img.max() > max_val:
        new_nbit = int(np.ceil(np.log2(img.max())))
        print(f"Detected pixel value larger than {nbits}-bit range, using {new_nbit}-bit range.")
        max_val = 2**new_nbit - 1
    return max_val


def bayer2rgb_cc(
    img,
    nbits,
    down=None,
    blue_gain=None,
    red_gain=None,
    black_level=RPI_HQ_CAMERA_BLACK_LEVEL,
    ccm=RPI_HQ_CAMERA_CCM_MATRIX,
    nbits_out=None,
):
    """
    Convert raw Bayer data to RGB with the following steps:

    #. Demosaic with bi-linear interpolation, mapping the Bayer array to RGB.
    #. Black level removal.
    #. White balancing, applying gains to red and blue channels.
    #. Color correction matrix.
    #. Clip.

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray`
        2D Bayer data to convert to RGB.
    nbits : int
        Bit depth of input data.
    blue_gain : float
        Blue gain.
    red_gain : float
        Red gain.
    black_level : float
        Black level. Default is to use that of Raspberry Pi HQ camera.
    ccm : :py:class:`~numpy.ndarray`
        Color correction matrix. Default is to use that of Raspberry Pi HQ camera.
    nbits_out : int
        Output bit depth. Default is to use that of input.

    Returns
    -------
    rgb : :py:class:`~numpy.ndarray`
        RGB data.
    """
    assert len(img.shape) == 2, img.shape
    if nbits_out is None:
        nbits_out = nbits
    if nbits_out > 8:
        dtype = np.uint16
    else:
        dtype = np.uint8

    # demosaic Bayer data
    img = cv2.cvtColor(img, cv2.COLOR_BayerRG2RGB)

    # downsample
    if down is not None:
        img = resize(img[None, ...], factor=1 / down, interpolation=cv2.INTER_CUBIC)[0]

    # correction
    img = img - black_level
    if red_gain:
        img[:, :, 0] *= red_gain
    if blue_gain:
        img[:, :, 2] *= blue_gain
    img = img / (2**nbits - 1 - black_level)
    img[img > 1] = 1

    img = (img.reshape(-1, 3, order="F") @ ccm.T).reshape(img.shape, order="F")
    img[img < 0] = 0
    img[img > 1] = 1
    return (img * (2**nbits_out - 1)).astype(dtype)


def print_image_info(img):
    """
    Print dimensions, data type, max, min, mean.
    """
    print("dimensions : {}".format(img.shape))
    print("data type : {}".format(img.dtype))
    print("max  : {}".format(img.max()))
    print("min  : {}".format(img.min()))
    print("mean : {}".format(img.mean()))


def autocorr2d(vals, pad_mode="reflect"):
    """
    Compute 2-D autocorrelation of image via the FFT.

    Parameters
    ----------
    vals : :py:class:`~numpy.ndarray`
        2-D image.
    pad_mode : str
        Desired padding. See NumPy documentation: https://numpy.org/doc/stable/reference/generated/numpy.pad.html

    Return
    ------
    autocorr : :py:class:`~numpy.ndarray`
    """

    shape = vals.shape
    assert len(shape) == 2

    # pad
    vals_padded = np.pad(
        vals,
        pad_width=((shape[0] // 2, shape[0] // 2), (shape[1] // 2, shape[1] // 2)),
        mode=pad_mode,
    )

    # compute autocorrelation via DFT
    X = np.fft.rfft2(vals_padded)
    autocorr = np.fft.ifftshift(np.fft.irfft2(X * X.conj()))

    # remove padding
    return autocorr[shape[0] // 2 : -shape[0] // 2, shape[1] // 2 : -shape[1] // 2]


def corr2d(im1, im2):
    """
    Source: https://stackoverflow.com/a/24769222

    """

    # get rid of the color channels by performing a grayscale transform
    # the type cast into 'float' is to avoid overflows
    im1_gray = np.sum(im1.astype("float"), axis=2)
    im2_gray = np.sum(im2.astype("float"), axis=2)

    # get rid of the averages, otherwise the results are not good
    im1_gray -= np.mean(im1_gray)
    im2_gray -= np.mean(im2_gray)

    # calculate the correlation image; note the flipping of onw of the images
    return scipy.signal.fftconvolve(im1_gray, im2_gray[::-1, ::-1], mode="same")


def rgb2bayer(img, pattern):
    """
    Converting RGB image to separated Bayer channels.

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray`
        Image in RGB format.
    pattern : str
        Bayer pattern: `RGGB`, `BGGR`, `GRBG`, `GBRG`.

    Returns
    -------
    :py:class:`~numpy.ndarray`
        Image converted to the Bayer format `[R, Gr, Gb, B]`. `Gr` and `Gb` are for the green pixels that are on the same line as the red and blue pixels respectively.
    """

    # Verifying that the pattern is a proper Bayer pattern
    pattern = pattern.upper()
    assert pattern in [
        "RGGB",
        "BGGR",
        "GRBG",
        "GBRG",
    ], "Bayer pattern must be in ['RGGB', 'BGGR', 'GRBG', 'GBRG']"

    # Doubling the size of the image to anticipatie shrinking from Bayer transformation
    height, width, _ = img.shape
    resized = resize(img, shape=(height * 2, width * 2, 3))

    # Separating each Bayer channel

    if pattern == "RGGB":
        # RGGB pattern *------*
        #              | R  G |
        #              | G  B |
        #              *------*
        r = resized[::2, ::2, 0]
        gr = resized[1::2, ::2, 1]
        gb = resized[::2, 1::2, 1]
        b = resized[1::2, 1::2, 2]
        img_bayer = np.dstack((r, gr, gb, b))

    elif pattern == "BGGR":
        # BGGR pattern *------*
        #              | B  G |
        #              | G  R |
        #              *------*
        r = resized[1::2, 1::2, 0]
        gr = resized[::2, 1::2, 1]
        gb = resized[1::2, ::2, 1]
        b = resized[::2, ::2, 2]
        img_bayer = np.dstack((b, gb, gr, r))

    elif pattern == "GRBG":
        # GRGB pattern *------*
        #              | G  R |
        #              | B  G |
        #              *------*
        r = resized[1::2, ::2, 0]
        gr = resized[::2, ::2, 1]
        gb = resized[1::2, 1::2, 1]
        b = resized[::2, 1::2, 2]
        img_bayer = np.dstack((gr, r, b, gb))

    else:
        # GBRG pattern *------*
        #              | G  B |
        #              | R  G |
        #              *------*
        r = resized[::2, 1::2, 0]
        gr = resized[1::2, 1::2, 1]
        gb = resized[::2, ::2, 1]
        b = resized[1::2, ::2, 2]
        img_bayer = np.dstack((gb, b, r, gr))

    return img_bayer


def bayer2rgb(X_bayer, pattern):
    """
    Converting 4-channel Bayer image to RGB by averaging the two green channels.

    Parameters
    ----------
    X_bayer : :py:class:`~numpy.ndarray`
        Image in RGB format.
    pattern : str
        Bayer pattern: `RGGB`, `BGGR`, `GRBG`, `GBRG`.

    Returns
    -------
    :py:class:`~numpy.ndarray`
        Image converted to the RGB format.
    """

    # Verifying that the pattern is a proper Bayer pattern
    pattern = pattern.upper()
    assert pattern in [
        "RGGB",
        "BGGR",
        "GRBG",
        "GBRG",
    ], "Bayer pattern must be in ['RGGB', 'BGGR', 'GRBG', 'GBRG']"

    r_channel = [i for i, ltr in enumerate(pattern) if ltr == "R"][0]
    b_channel = [i for i, ltr in enumerate(pattern) if ltr == "B"][0]
    g_channels = [i for i, ltr in enumerate(pattern) if ltr == "G"]

    X_rgb = np.empty(X_bayer.shape[:-1] + (3,))
    X_rgb[:, :, 0] = X_bayer[:, :, r_channel]
    X_rgb[:, :, 1] = np.mean(X_bayer[:, :, g_channels], axis=2)
    X_rgb[:, :, 2] = X_bayer[:, :, b_channel]

    return X_rgb



"""
Prepare a subset of the DiffuserCam dataset: https://github.com/Waller-Lab/LenslessLearning
which also includes viewable/postprocessed images.

Note that you need the dataset locally, which is about 100 GB.

```bash
python scripts/data/prepare_mirflickr_subset.py \
--data ~/Documents/DiffuserCam/DiffuserCam_Mirflickr_Dataset
```

"""

import click
import glob
import numpy as np
from datetime import datetime
import os
import pathlib as plib
from shutil import copyfile
from PIL import Image
from lensless.recon.mirflickr import postprocess


@click.command()
@click.option(
    "--data",
    type=str,
    help="Path to original dataset.",
)
@click.option(
    "--n_files",
    type=int,
    default=200,
    help="Number of files in subset.",
)
@click.option(
    "--seed",
    type=int,
    default=11,
)
@click.option(
    "--output_dir_path",
    type=str,
    help="Output directory. Otherwise save subset in running directory.",
)
def subset_mirflickr(data, n_files, seed, output_dir_path):
    diffuser_dir = os.path.join(os.path.join(data, "dataset"), "diffuser_images")
    lensed_dir = os.path.join(os.path.join(data, "dataset"), "ground_truth_lensed")
    psf_path = os.path.join(data, "psf.tiff")
    output_dir_path = None

    # create output directory
    timestamp = datetime.now().strftime("%d%m%Y_%Hh%M")
    output_dir_fn = f"DiffuserCam_Mirflickr_{n_files}_{timestamp}_seed{seed}"
    if output_dir_path is not None:
        assert os.path.isdir(output_dir_path)
        output_dir = os.path.join(output_dir_path, output_dir_fn)
    else:
        output_dir = output_dir_fn
    output_plib = plib.Path(output_dir)
    output_plib.mkdir(exist_ok=False)
    diffuser_out = os.path.join(output_dir, "diffuser")
    os.makedirs(diffuser_out)
    lensed_out = os.path.join(output_dir, "lensed")
    os.makedirs(lensed_out)
    print(f"Created output directory : {output_dir}")

    # shuffle and take first few
    diffuser_files = glob.glob(diffuser_dir + "/*.npy")
    np.random.seed(seed)
    np.random.shuffle(diffuser_files)
    subset = diffuser_files[:n_files]

    # copy over files
    copyfile(psf_path, os.path.join(output_dir, os.path.basename(psf_path)))
    for fn in subset:
        bn = os.path.basename(fn)

        # copy raw data and viewable format
        copyfile(fn, os.path.join(diffuser_out, bn))
        image_data = postprocess(np.load(fn))
        image_data = (image_data * 255).astype(np.uint8)
        viewable_file = os.path.join(diffuser_out, bn.split(".")[0] + ".tif")
        im = Image.fromarray(image_data)
        im.save(viewable_file)

        # copy lensed data and viewable format
        lensed_fp = os.path.join(lensed_dir, bn)
        copyfile(lensed_fp, os.path.join(lensed_out, bn))
        image_data = (postprocess(np.load(lensed_fp)) * 255).astype(np.uint8)
        viewable_file = os.path.join(lensed_out, bn.split(".")[0] + ".tif")
        im = Image.fromarray(image_data)
        im.save(viewable_file)


if __name__ == "__main__":
    subset_mirflickr()



"""
Apply Accelerated Proximal Gradient Descent (APGD) with a desired prior.

Pycsou documentation of (APGD):
https://matthieumeo.github.io/pycsou/html/api/algorithms/pycsou.opt.proxalgs.html?highlight=apgd#pycsou.opt.proxalgs.AcceleratedProximalGradientDescent

Example (default to non-negativity prior):
```
python scripts/recon/apgd_pycsou.py
```

"""

import hydra
from hydra.utils import to_absolute_path
import numpy as np
import time
import matplotlib.pyplot as plt
from lensless.utils.io import load_data
from lensless.recon.apgd import APGD
import os
import pathlib as plib


import logging

# A logger for this file
log = logging.getLogger(__name__)


@hydra.main(version_base=None, config_path="../../configs/recon", config_name="apgd_l1")
def apgd(
    config,
):
    psf, data = load_data(
        psf_fp=to_absolute_path(config["input"]["psf"]),
        data_fp=to_absolute_path(config["input"]["data"]),
        dtype=config.input.dtype,
        downsample=config["preprocess"]["downsample"],
        bayer=config["preprocess"]["bayer"],
        blue_gain=config["preprocess"]["blue_gain"],
        red_gain=config["preprocess"]["red_gain"],
        plot=config["display"]["plot"],
        flip=config["preprocess"]["flip"],
        gamma=config["display"]["gamma"],
        gray=config["preprocess"]["gray"],
        single_psf=config["preprocess"]["single_psf"],
        shape=config["preprocess"]["shape"],
    )

    disp = config["display"]["disp"]
    if disp < 0:
        disp = None

    save = config["save"]
    if save:
        save = os.getcwd()

    start_time = time.time()
    recon = APGD(psf=psf, **config.apgd)
    recon.set_data(data)
    print(f"Setup time : {time.time() - start_time} s")

    start_time = time.time()
    res = recon.apply(
        n_iter=config["apgd"]["max_iter"],
        disp_iter=disp,
        save=save,
        gamma=config["display"]["gamma"],
        plot=config["display"]["plot"],
    )
    print(f"Processing time : {time.time() - start_time} s")
    final_img = res[0]

    if config["display"]["plot"]:
        plt.show()
    if save:
        np.save(plib.Path(save) / "final_reconstruction.npy", final_img)
        print(f"Files saved to : {save}")


if __name__ == "__main__":
    apgd()



# #############################################################################
# mirflickr.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


import numpy as np
from lensless.recon.admm import ADMM
import cv2


class ADMM_MIRFLICKR(ADMM):
    """
    Object for applying ADMM (Alternating Direction Method of Multipliers) to
    the DiffuserCam Lensless Mirflickr Dataset.

    New class as a unique `_form_image` function has to be defined to crop the
    data at the suitable region.

    Original dataset page: https://waller-lab.github.io/LenslessLearning/dataset.html
    Subset (200 files) we've prepared: https://drive.switch.ch/index.php/s/vmAZzryGI8U8rcE

    """

    def __init__(self, *args, **kwargs):
        # call reset() to initialize matrices
        super(ADMM_MIRFLICKR, self).__init__(*args, **kwargs)

    def _form_image(self):
        image = self._crop(self._image_est)
        image[image < 0] = 0
        return postprocess(image)


def postprocess(image):
    # https://github.com/Waller-Lab/LenslessLearning/blob/eaab0fb694a4f51fdda382a53f92832af98fd692/utils.py#L429
    image = image.astype(np.float32)
    image_color = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    out_image = np.flipud(np.clip(image_color, 0, 1))
    return out_image[60:, 62:-38, :]



"""
First key is camera, second key is training data, third key is model name.

Download link corresponds to output folder from training
script, which contains the model checkpoint and config file,
and other intermediate files. Models are stored on Hugging Face.
"""


import os
from lensless.recon.integrated_background_sub import IntegratedBackgroundSub
import numpy as np
import torch
from lensless.recon.utils import create_process_network
from lensless.recon.unrolled_admm import UnrolledADMM
from lensless.recon.trainable_inversion import TrainableInversion
from lensless.hardware.trainable_mask import prep_trainable_mask
import yaml
from lensless.recon.multi_wiener import MultiWiener
from lensless import SVDeconvNet
from huggingface_hub import snapshot_download
from collections import OrderedDict
from lensless.utils.dataset import MyDataParallel


model_dir_path = os.path.join(os.path.dirname(__file__), "..", "..", "models")

model_dict = {
    "diffusercam": {
        "mirflickr": {
            # -- only unrolled20
            "U20": "bezzam/diffusercam-mirflickr-unrolled-admm20",
            "U20_0db": "bezzam/diffusercam-mirflickr-unrolled-admm20-0db",
            "U20_10db": "bezzam/diffusercam-mirflickr-unrolled-admm20-10db",
            "U20_20db": "bezzam/diffusercam-mirflickr-unrolled-admm20-20db",
            # -- only pre-process
            "Unet+U20": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20",
            "Unet+U20_0dB": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-0db",
            "Unet+U20_10db": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-10db",
            "Unet+U20_20db": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-20db",
            # -- only post-process
            "U20+Unet": "bezzam/diffusercam-mirflickr-unrolled-admm20-unet2",
            "U20+Unet_0db": "bezzam/diffusercam-mirflickr-unrolled-admm20-unet2-0db",
            "U20+Unet_10db": "bezzam/diffusercam-mirflickr-unrolled-admm20-unet2-10db",
            "U20+Unet_20db": "bezzam/diffusercam-mirflickr-unrolled-admm20-unet2-20db",
            # -- only post-process (Drunet)
            "U20+Drunet": "bezzam/diffusercam-mirflickr-unrolled-admm20-drunet",
            "TrainInv+Drunet": "bezzam/diffusercam-mirflickr-trainable-inv-drunet",
            # -- both
            "Unet+TrainInv+Unet": "bezzam/diffusercam-mirflickr-unet2-trainable-inv-unet2",
            "Unet+U20+Unet": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-unet2",  # init with 0.01 and only pre-proc
            "Unet+U20+Unet_aux0.01": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-unet2-aux0.01",
            "Unet+U20+Unet_aux0.03": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-unet2-aux0.03",
            "Unet+U20+Unet_aux0.1": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-unet2-aux0.1",  # init with 0.01
            "Unet+U20+Unet_aux1": "bezzam/diffusercam-mirflickr-unet2-unrolled-admm20-unet2-aux1",  # init with 0.01 and only pre-proc
            # baseline benchmarks which don't have model file but use ADMM
            "admm_fista": "bezzam/diffusercam-mirflickr-admm-fista",
            "admm_pnp": "bezzam/diffusercam-mirflickr-admm-pnp",
            # -- TCI submission
            "TrainInv+Unet8M": "bezzam/diffusercam-mirflickr-trainable-inv-unet8M",
            "Unet4M+U5+Unet4M": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M",
            "MWDN8M": "bezzam/diffusercam-mirflickr-mwdn-8M",
            "Unet2M+MWDN6M": "bezzam/diffusercam-mirflickr-unet2M-mwdn-6M",
            "Unet4M+TrainInv+Unet4M": "bezzam/diffusercam-mirflickr-unet4M-trainable-inv-unet4M",
            "MMCN4M+Unet4M": "bezzam/diffusercam-mirflickr-mmcn-unet4M",
            "U5+Unet8M": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M",
            "Unet8M+U5": "bezzam/diffusercam-mirflickr-unet8M-unrolled-admm5",
            "Unet2M+MMCN+Unet2M": "bezzam/diffusercam-mirflickr-unet2M-mmcn-unet2M",
            "Unet4M+U20+Unet4M": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm20-unet4M",
            "Unet4M+U10+Unet4M": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm10-unet4M",
            "Unet4M+U5+Unet4M_psfNN": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psfNN",
            # training with PSF noise
            "U5+Unet8M_psf0dB": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M-psf0dB",
            "U5+Unet8M_psf-5dB": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M-psf-5dB",
            "U5+Unet8M_psf-10dB": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M-psf-10dB",
            "U5+Unet8M_psf-20dB": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M-psf-20dB",
            "Unet4M+U5+Unet4M_psf-0dB": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psf-0dB",
            "Unet4M+U5+Unet4M_psf-5dB": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psf-5dB",
            "Unet4M+U5+Unet4M_psf-10dB": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psf-10dB",
            "Unet4M+U5+Unet4M_psf-20dB": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psf-20dB",
            "Unet4M+U5+Unet4M_psfNN_psf-0dB": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psfNN-psf-0dB",
            "Unet4M+U5+Unet4M_psfNN_psf-10dB": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psfNN-psf-10dB",
            "Unet4M+U5+Unet4M_psfNN_psf-20dB": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psfNN-psf-20dB",
            # training with noise
            "U5+Unet8M_10db": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M-10db",
            "U5+Unet8M_40db": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M-40db",
            "Unet4M+U5+Unet4M_10db": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-10db",
            "Unet4M+U5+Unet4M_40db": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-40db",
            # fine-tuning tapecam
            "Unet4M+U5+Unet4M_ft_tapecam": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-ft-tapecam",
            "Unet4M+U5+Unet4M_ft_tapecam_post": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-ft-tapecam-post",
            "Unet4M+U5+Unet4M_ft_tapecam_pre": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-ft-tapecam-pre",
            # comparing with transformers, with ADAMW optimizer (below with ADAMW optimizer)
            "U5+Unet8M_adamw": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet8M-adamw",
            "Unet4M+U5+Unet4M_adamw": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-adamw",
            "Unet4M+U5+Unet4M_psfNN_adamw": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psfNN-adamw",
            "U5+Transformer8M": "bezzam/diffusercam-mirflickr-unrolled-admm5-transformer8M",
            "Transformer4M+U5+Transformer4M": "bezzam/diffusercam-mirflickr-transformer4M-unrolled-admm5-transformer4M",
            "Transformer4M+U5+Transformer4M_psfNN": "bezzam/difusercam-mirflickr-transformer4M-unrolled-admm5-transformer4M-psfNN",
            # (~11.6M param) comparing with SVDeconvNet, with ADAMW optimizer (full resolution images)
            "U5+Unet12M_fullres": "bezzam/diffusercam-mirflickr-unrolled-admm5-unet12M-fullres",
            "Unet6M+U5+Unet6M_fullres": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-fullres",
            "Unet6M+U5+Unet6M_psfNN_fullres": "bezzam/diffusercam-mirflickr-unet4M-unrolled-admm5-unet4M-psfNN-fullres",
            "SVDecon+UNet8M": "bezzam/diffusercam-mirflickr-svdecon-unet4M",
            "Unet4M+SVDecon+Unet4M": "bezzam/diffusercam-mirflickr-unet4M-svdecon-unet4M",
        },
        "mirflickr_sim": {
            "Unet4M+U5+Unet4M": "bezzam/diffusercam-mirflickr-sim-unet4M-unrolled-admm5-unet4M",
            "Unet4M+U5+Unet4M_ft_tapecam": "bezzam/diffusercam-mirflickr-sim-unet4M-unrolled-admm5-unet4M-ft-tapecam",
            "Unet4M+U5+Unet4M_ft_tapecam_post": "bezzam/diffusercam-mirflickr-sim-unet4M-unrolled-admm5-unet4M-ft-tapecam-post",
            "Unet4M+U5+Unet4M_ft_tapecam_pre": "bezzam/diffusercam-mirflickr-sim-unet4M-unrolled-admm5-unet4M-ft-tapecam-pre",
            "Unet4M+U5+Unet4M_ft_digicam_multi_post": "bezzam/diffusercam-mirflickr-sim-unet4M-unrolled-admm5-unet4M-ft-digicam-multi-post",
            "Unet4M+U5+Unet4M_ft_digicam_multi_pre": "bezzam/diffusercam-mirflickr-sim-unet4M-unrolled-admm5-unet4M-ft-digicam-multi-pre",
            "Unet4M+U5+Unet4M_ft_digicam_multi": "bezzam/diffusercam-mirflickr-sim-unet4M-unrolled-admm5-unet4M-ft-digicam-multi",
        },
    },
    "digicam": {
        "celeba_26k": {
            "unrolled_admm10": "bezzam/digicam-celeba-unrolled-admm10",
            "unrolled_admm10_ft_psf": "bezzam/digicam-celeba-unrolled-admm10-ft-psf",
            "unet8M": "bezzam/digicam-celeba-unet8M",
            "TrainInv+Unet8M": "bezzam/digicam-celeba-trainable-inv-unet8M",
            "unrolled_admm10_post8M": "bezzam/digicam-celeba-unrolled-admm10-post8M",
            "unrolled_admm10_ft_psf_post8M": "bezzam/digicam-celeba-unrolled-admm10-ft-psf-post8M",
            "pre8M_unrolled_admm10": "bezzam/digicam-celeba-pre8M-unrolled-admm10",
            "pre4M_unrolled_admm10_post4M": "bezzam/digicam-celeba-pre4M-unrolled-admm10-post4M",
            "pre4M_unrolled_admm10_ft_psf_post4M": "bezzam/digicam-celeba-pre4M-unrolled-admm10-ft-psf-post4M",
            "Unet4M+TrainInv+Unet4M": "bezzam/digicam-celeba-unet4M-trainable-inv-unet4M",
            # baseline benchmarks which don't have model file but use ADMM
            "admm_measured_psf": "bezzam/digicam-celeba-admm-measured-psf",
            "admm_simulated_psf": "bezzam/digicam-celeba-admm-simulated-psf",
            # TCI submission (using waveprop simulation)
            "U5+Unet8M_wave": "bezzam/digicam-celeba-unrolled-admm5-unet8M",
            "Unet8M+U5_wave": "bezzam/digicam-celeba-unet8M-unrolled-admm5",
            "TrainInv+Unet8M_wave": "bezzam/digicam-celeba-trainable-inv-unet8M_wave",
            "MWDN8M_wave": "bezzam/digicam-celeba-mwnn-8M",
            "MMCN4M+Unet4M_wave": "bezzam/digicam-celeba-mmcn-unet4M",
            "Unet2M+MWDN6M_wave": "bezzam/digicam-celeba-unet2M-mwdn-6M",
            "Unet4M+TrainInv+Unet4M_wave": "bezzam/digicam-celeba-unet4M-trainable-inv-unet4M_wave",
            "Unet2M+MMCN+Unet2M_wave": "bezzam/digicam-celeba-unet2M-mmcn-unet2M",
            "Unet4M+U5+Unet4M_wave": "bezzam/digicam-celeba-unet4M-unrolled-admm5-unet4M",
            "Unet4M+U10+Unet4M_wave": "bezzam/digicam-celeba-unet4M-unrolled-admm10-unet4M",
            "Unet4M+U5+Unet4M_wave_psfNN": "bezzam/digicam-celeba-unet4M-unrolled-admm5-unet4M-wave-psfNN",
        },
        "mirflickr_single_25k": {
            # simulated PSF (without waveprop, with deadspace)
            "U10": "bezzam/digicam-mirflickr-single-25k-unrolled-admm10",
            "Unet8M": "bezzam/digicam-mirflickr-single-25k-unet8M",
            "TrainInv+Unet8M": "bezzam/digicam-mirflickr-single-25k-trainable-inv-unet8M",
            "U10+Unet8M": "bezzam/digicam-mirflickr-single-25k-unrolled-admm10-unet8M",
            "Unet4M+TrainInv+Unet4M": "bezzam/digicam-mirflickr-single-25k-unet4M-trainable-inv-unet4M",
            "Unet4M+U10+Unet4M": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm10-unet4M",
            # simulated PSF (with waveprop, with deadspace)
            "U10_wave": "bezzam/digicam-mirflickr-single-25k-unrolled-admm10-wave",
            "U10+Unet8M_wave": "bezzam/digicam-mirflickr-single-25k-unrolled-admm10-unet8M-wave",
            "Unet8M_wave": "bezzam/digicam-mirflickr-single-25k-unet8M-wave",
            "Unet8M_wave_v2": "bezzam/digicam-mirflickr-single-25k-unet8M-wave-v2",
            "Unet4M+U10+Unet4M_wave": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm10-unet4M-wave",
            "TrainInv+Unet8M_wave": "bezzam/digicam-mirflickr-single-25k-trainable-inv-unet8M-wave",
            "U5+Unet8M_wave": "bezzam/digicam-mirflickr-single-25k-unrolled-admm5-unet8M-wave",
            "Unet8M+U5_wave": "bezzam/digicam-mirflickr-single-25k-unet8M-unrolled-admm5-wave",
            "Unet4M+U5+Unet4M_wave": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm5-unet4M-wave",
            "Unet4M+U5+Unet4M_wave_psfNN": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm5-unet4M-wave-psfNN",
            "MWDN8M_wave": "bezzam/digicam-mirflickr-single-25k-mwdn-8M",
            "MMCN4M+Unet4M_wave": "bezzam/digicam-mirflickr-single-25k-mmcn-unet4M",
            "Unet2M+MMCN+Unet2M_wave": "bezzam/digicam-mirflickr-single-25k-unet2M-mmcn-unet2M-wave",
            "Unet4M+TrainInv+Unet4M_wave": "bezzam/digicam-mirflickr-single-25k-unet4M-trainable-inv-unet4M-wave",
            "Unet2M+MWDN6M_wave": "bezzam/digicam-mirflickr-single-25k-unet2M-mwdn-6M",
            "Unet4M+U5+Unet4M_wave_aux1": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm5-unet4M-wave-aux1",
            "Unet4M+U5+Unet4M_wave_flips": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm5-unet4M-wave-flips",
            "Unet4M+U5+Unet4M_wave_flips_rotate10": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm5-unet4M-wave-flips-rotate10",
            # measured PSF
            "Unet4M+U10+Unet4M_measured": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm10-unet4M-measured",
            # simulated PSF (with waveprop, no deadspace)
            "Unet4M+U10+Unet4M_wave_nodead": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm10-unet4M-wave-nodead",
            # simulated PSF (without waveprop, no deadspace)
            "Unet4M+U10+Unet4M_nodead": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm10-unet4M-nodead",
            # finetune
            "Unet4M+U5+Unet4M_ft_flips": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm5-unet4M-ft-flips",
            "Unet4M+U5+Unet4M_ft_flips_rotate10": "bezzam/digicam-mirflickr-single-25k-unet4M-unrolled-admm5-unet4M-ft-flips-rotate10",
        },
        "mirflickr_multi_25k": {
            # simulated PSFs (without waveprop, with deadspace)
            "Unet8M": "bezzam/digicam-mirflickr-multi-25k-unet8M",
            "Unet8M_wave_v2": "bezzam/digicam-mirflickr-multi-25k-unet8M-wave-v2",
            "Unet4M+U10+Unet4M": "bezzam/digicam-mirflickr-multi-25k-unet4M-unrolled-admm10-unet4M",
            # simulated PSF (with waveprop, with deadspace)
            "Unet4M+U10+Unet4M_wave": "bezzam/digicam-mirflickr-multi-25k-unet4M-unrolled-admm10-unet4M-wave",
            "Unet4M+U5+Unet4M_wave": "bezzam/digicam-mirflickr-multi-25k-unet4M-unrolled-admm5-unet4M-wave",
            "Unet4M+U5+Unet4M_wave_psfNN": "bezzam/digicam-mirflickr-multi-25k-unet4M-unrolled-admm5-unet4M-wave-psfNN",
            "Unet4M+U5+Unet4M_wave_aux1": "bezzam/digicam-mirflickr-multi-25k-unet4M-unrolled-admm5-unet4M-wave-aux1",
            "Unet4M+U5+Unet4M_wave_flips": "bezzam/digicam-mirflickr-multi-25k-unet4M-unrolled-admm5-unet4M-wave-flips",
        },
    },
    "tapecam": {
        "mirflickr": {
            "U5+Unet8M": "bezzam/tapecam-mirflickr-unrolled-admm5-unet8M",
            "Unet8M+U5": "bezzam/tapecam-mirflickr-unet8M-unrolled-admm5",
            "TrainInv+Unet8M": "bezzam/tapecam-mirflickr-trainable-inv-unet8M",
            "MMCN4M+Unet4M": "bezzam/tapecam-mirflickr-mmcn-unet4M",
            "MWDN8M": "bezzam/tapecam-mirflickr-mwdn-8M",
            "Unet4M+TrainInv+Unet4M": "bezzam/tapecam-mirflickr-unet4M-trainable-inv-unet4M",
            "Unet4M+U5+Unet4M": "bezzam/tapecam-mirflickr-unet4M-unrolled-admm5-unet4M",
            "Unet2M+MMCN+Unet2M": "bezzam/tapecam-mirflickr-unet2M-mmcn-unet2M",
            "Unet2M+MWDN6M": "bezzam/tapecam-mirflickr-unet2M-mwdn-6M",
            "Unet4M+U10+Unet4M": "bezzam/tapecam-mirflickr-unet4M-unrolled-admm10-unet4M",
            "Unet4M+U5+Unet4M_flips": "bezzam/tapecam-mirflickr-unet4M-unrolled-admm5-unet4M-flips",
            "Unet4M+U5+Unet4M_flips_rotate10": "bezzam/tapecam-mirflickr-unet4M-unrolled-admm5-unet4M-flips-rotate10",
            "Unet4M+U5+Unet4M_aux1": "bezzam/tapecam-mirflickr-unet4M-unrolled-admm5-unet4M-aux1",
            "Unet4M+U5+Unet4M_psfNN": "bezzam/tapecam-mirflickr-unet4M-unrolled-admm5-unet4M-psfNN",
        },
    },
    "multilens": {
        "mirflickr_ambient": {
            "U5+Unet8M": "lensless/multilens-mirflickr-ambient-unrolled-admm5-unet8M",
            "U5+Unet8M_direct_sub": "lensless/multilens-mirflickr-ambient-unrolled-admm5-unet8M-direct-sub",
            "U5+Unet8M_learned_sub": "lensless/multilens-mirflickr-ambient-unrolled-admm5-unet8M-learned-sub",
            "Unet4M+U5+Unet4M": "lensless/multilens-mirflickr-ambient-unet4M-unrolled-admm5-unet4M",
            "Unet4M+U5+Unet4M_direct_sub": "lensless/multilens-mirflickr-ambient-unet4M-unrolled-admm5-unet4M-direct-sub",
            "Unet4M+U5+Unet4M_learned_sub": "lensless/multilens-mirflickr-ambient-unet4M-unrolled-admm5-unet4M-learned-sub",
            "Unet4M+U5+Unet4M_concat": "lensless/multilens-mirflickr-ambient-unet4M-unrolled-admm5-unet4M-concat-ext",
            "Unet4M+U5+Unet4M_concat_psfNN": "lensless/multilens-mirflickr-ambient-unet4M-unrolled-admm5-unet4M-concat-psfNN",
            "TrainInv+Unet8M": "lensless/multilens-mirflickr-ambient-trainable-inv-unet8M",
            "TrainInv+Unet8M_learned_sub": "lensless/multilens-mirflickr-ambient-trainable-inv-unet8M-learned-sub",
            "Unet4M+TrainInv+Unet4M": "lensless/multilens-mirflickr-ambient-unet4M-trainable-inv-unet4M",
            "Unet4M+TrainInv+Unet4M_learned_sub": "lensless/multilens-mirflickr-ambient-unet4M-trainable-inv-unet4M-learned-sub",
            "Unet4M+TrainInv+Unet4M_concat": "lensless/multilens-mirflickr-ambient-unet4M-trainable-inv-unet4M-concat-ext",
            "TrainInv+Unet8M_direct_sub": "lensless/multilens-mirflickr-ambient-trainable-inv-unet8M-direct-sub",
            "Unet4M+TrainInv+Unet4M_direct_sub": "lensless/multilens-mirflickr-ambient-unet4M-trainable-inv-unet4M-direct-sub",
        }
    },
}


def remove_data_parallel(old_state_dict):
    new_state_dict = OrderedDict()

    for k, v in old_state_dict.items():
        # remove `module.` from k
        if "module." in k:
            name = k.replace("module.", "")
            # name = k[7:] # remove `module.`
            new_state_dict[name] = v

    return new_state_dict


def download_model(camera, dataset, model, local_model_dir=None):

    """
    Download model from model_dict (if needed).

    Parameters
    ----------
    dataset : str
        Dataset used for training.
    model_name : str
        Name of model.
    """

    if local_model_dir is None:
        local_model_dir = model_dir_path

    if camera not in model_dict:
        raise ValueError(f"Camera {camera} not found in model_dict.")

    if dataset not in model_dict[camera]:
        raise ValueError(f"Dataset {dataset} not found in model_dict.")

    if model not in model_dict[camera][dataset]:
        raise ValueError(f"Model {model} not found in model_dict.")

    repo_id = model_dict[camera][dataset][model]
    model_dir = os.path.join(local_model_dir, camera, dataset, model)

    if not os.path.exists(model_dir):
        snapshot_download(repo_id=repo_id, local_dir=model_dir)

    return model_dir


def load_model(
    model_path,
    psf,
    device="cpu",
    device_ids=None,
    legacy_denoiser=False,
    verbose=True,
    skip_pre=False,
    skip_post=False,
    train_last_layer=False,
    return_intermediate=False,
):

    """
    Load best model from model path.

    Parameters
    ----------
    model_path : str
        Path to model.
    psf : py:class:`~torch.Tensor`
        PSF tensor.
    device : str
        Device to load model on.
    """

    # load Hydra config
    config_path = os.path.join(model_path, ".hydra", "config.yaml")
    with open(config_path, "r") as stream:
        config = yaml.safe_load(stream)

    # load learning mask
    downsample = config["files"]["downsample"] * 4  # TODO: particular to DiffuserCam?
    mask = prep_trainable_mask(config, psf, downsample=downsample)
    if mask is not None:

        if config["trainable_mask"]["mask_type"] == "TrainablePSF":
            # print("loading best PSF...")

            psf_learned = np.load(os.path.join(model_path, "psf_epochBEST.npy"))
            psf_learned = torch.Tensor(psf_learned).to(psf).unsqueeze(0)

            # -- set values and get new PSF
            with torch.no_grad():
                mask._mask = torch.nn.Parameter(torch.tensor(psf_learned))
                psf = mask.get_psf().to(device)

    # load best model config
    model_checkpoint = os.path.join(model_path, "recon_epochBEST")
    assert os.path.exists(model_checkpoint), "Checkpoint does not exist"
    if verbose:
        print("Loading checkpoint from : ", model_checkpoint)
    model_state_dict = torch.load(model_checkpoint, map_location=device)

    if config["files"].get("psf_snr", None) is not None:
        # overwrite PSF with noisy PSF used during training
        psf_path = os.path.join(model_path, "psf.pt")
        assert os.path.exists(psf_path), "PSF does not exist"
        psf = torch.load(psf_path, map_location=device)

    # load model
    pre_process = None
    post_process = None

    background_network = None
    if config["reconstruction"].get("direct_background_subtraction", False):
        assert config["reconstruction"].get("learned_background_subtraction", None) is False
    elif config["reconstruction"].get("learned_background_subtraction", False):
        assert config["reconstruction"].get("direct_background_subtraction", None) is False

        # create UnetRes for background subtraction
        background_network, _ = create_process_network(
            network="UnetRes",
            depth=len(config["reconstruction"]["learned_background_subtraction"]),
            nc=config["reconstruction"]["learned_background_subtraction"],
            device=device,
            device_ids=device_ids,
        )
    elif config["reconstruction"].get("integrated_background_subtraction", False):
        assert config["reconstruction"]["direct_background_subtraction"] is False
        assert config["reconstruction"]["learned_background_subtraction"] is False
        assert pre_process is None

        background_network = IntegratedBackgroundSub(
            in_channels=3,
            out_channels=3,
            input_shape=psf.shape,
            nc=config["reconstruction"]["integrated_background_subtraction"],
            down_subtraction=config["reconstruction"]["down_subtraction"],
        )

    # network for PSF
    psf_network = None
    if config["reconstruction"].get("psf_network", None):
        # create UnetRes for PSF
        psf_network, _ = create_process_network(
            network="UnetRes",
            depth=len(config["reconstruction"]["psf_network"]),
            nc=config["reconstruction"]["psf_network"],
            device=device,
            device_ids=device_ids,
        )

    if config["reconstruction"].get("init", None):

        init_model = config["reconstruction"]["init"]
        assert config["reconstruction"].get("init_processors", None) is None

        param = init_model.split(":")
        assert len(param) == 4, "hf model requires following format: hf:camera:dataset:model_name"
        camera = param[1]
        dataset = param[2]
        model_name = param[3]
        model_path = download_model(camera=camera, dataset=dataset, model=model_name)
        recon = load_model(
            model_path,
            psf,
            device,
            device_ids=device_ids,
            train_last_layer=config["reconstruction"]["post_process"]["train_last_layer"],
        )

    else:

        if config["reconstruction"]["pre_process"]["network"] is not None:

            pre_process, _ = create_process_network(
                network=config["reconstruction"]["pre_process"]["network"],
                device=device,
                input_background=config["reconstruction"].get("unetres_input_background", False),
                # unetres param
                depth=config["reconstruction"]["pre_process"]["depth"],
                nc=config["reconstruction"]["pre_process"]["nc"]
                if "nc" in config["reconstruction"]["pre_process"].keys()
                else None,
                # restormer parameters
                restormer_params=config["reconstruction"]["pre_process"].get(
                    "restormer_params", None
                ),
            )

        if config["reconstruction"]["post_process"]["network"] is not None:

            post_process, _ = create_process_network(
                network=config["reconstruction"]["post_process"]["network"],
                device=device,
                # get from dict
                concatenate_compensation=config["reconstruction"]["compensation"][-1]
                if config["reconstruction"].get("compensation", None) is not None
                else False,
                # unetres param
                depth=config["reconstruction"]["post_process"]["depth"],
                nc=config["reconstruction"]["post_process"]["nc"]
                if "nc" in config["reconstruction"]["post_process"].keys()
                else None,
                # restormer parameters
                restormer_params=config["reconstruction"]["post_process"].get(
                    "restormer_params", None
                ),
            )

            if train_last_layer:
                for param in post_process.parameters():
                    for name, param in post_process.named_parameters():
                        if "m_tail" in name:
                            param.requires_grad = True
                        else:
                            param.requires_grad = False

        if config["reconstruction"]["method"] == "unrolled_admm":
            recon = UnrolledADMM(
                psf if mask is None else psf_learned,
                pre_process=pre_process,
                post_process=post_process,
                n_iter=config["reconstruction"]["unrolled_admm"]["n_iter"],
                skip_unrolled=config["reconstruction"]["skip_unrolled"],
                legacy_denoiser=legacy_denoiser,
                skip_pre=skip_pre,
                skip_post=skip_post,
                compensation=config["reconstruction"].get("compensation", None),
                psf_network=psf_network,
                psf_residual=config["reconstruction"].get("psf_residual", False),
                compensation_residual=config["reconstruction"].get("compensation_residual", False),
                direct_background_subtraction=config["reconstruction"].get(
                    "direct_background_subtraction", False
                ),
                integrated_background_subtraction=config["reconstruction"].get(
                    "integrated_background_subtraction", False
                ),
                background_network=background_network,
                return_intermediate=return_intermediate,
            )
        elif config["reconstruction"]["method"] == "trainable_inv":
            recon = TrainableInversion(
                psf,
                pre_process=pre_process,
                post_process=post_process,
                K=config["reconstruction"]["trainable_inv"]["K"],
                legacy_denoiser=legacy_denoiser,
                skip_pre=skip_pre,
                skip_post=skip_post,
                direct_background_subtraction=config["reconstruction"].get(
                    "direct_background_subtraction", False
                ),
                integrated_background_subtraction=config["reconstruction"].get(
                    "integrated_background_subtraction", False
                ),
                background_network=background_network,
                return_intermediate=return_intermediate,
            )
        elif config["reconstruction"]["method"] == "multi_wiener":

            if config["files"].get("single_channel_psf", False):

                if torch.sum(psf[..., 0] - psf[..., 1]) != 0:
                    # need to sum difference channels
                    raise ValueError("PSF channels are not the same")
                    # psf = np.sum(psf, axis=3)

                else:
                    psf = psf[..., 0].unsqueeze(-1)
                psf_channels = 1
            else:
                psf_channels = 3

            recon = MultiWiener(
                in_channels=3,
                out_channels=3,
                psf=psf,
                psf_channels=psf_channels,
                nc=config["reconstruction"]["multi_wiener"]["nc"],
                pre_process=pre_process,
            )
            recon.to(device)

        elif config["reconstruction"]["method"] == "svdeconvnet":
            psf_learned = psf_learned[0]  # remove singleton dimension
            recon = SVDeconvNet(
                psf[0].unsqueeze(
                    0
                ),  # set one of PSF (just for shape) but will be overwritten later
                K=config["reconstruction"]["svdeconvnet"]["K"],
                pre_process=pre_process,
                post_process=post_process,
                background_network=background_network,
                return_intermediate=return_intermediate,
                direct_background_subtraction=config["reconstruction"].get(
                    "direct_background_subtraction", False
                ),
                integrated_background_subtraction=config["reconstruction"].get(
                    "integrated_background_subtraction", False
                ),
            )

    if mask is not None:
        psf_learned = torch.nn.Parameter(psf_learned)
        recon._set_psf(psf_learned)

    if config.get("device_ids", None) is not None or isinstance(config.get("torch_device"), list):
        model_state_dict = remove_data_parallel(model_state_dict)

    # hotfixes for loading models
    if config["reconstruction"]["method"] == "multi_wiener":
        # replace "avgpool_conv" with "pool_conv"
        model_state_dict = {
            k.replace("avgpool_conv", "pool_conv"): v for k, v in model_state_dict.items()
        }

    recon.load_state_dict(model_state_dict)

    if device_ids is not None:
        if recon.pre_process is not None and background_network is None:
            pre_proc = torch.nn.DataParallel(recon.pre_process_model, device_ids=device_ids)
            pre_proc = pre_proc.to(device)
            recon.set_pre_process(pre_proc)
        if recon.post_process is not None:
            post_proc = torch.nn.DataParallel(recon.post_process_model, device_ids=device_ids)
            post_proc = post_proc.to(device)
            recon.set_post_process(post_proc)
        recon = MyDataParallel(recon, device_ids=device_ids)
    recon.to(device)
    return recon



from lensless.utils.io import load_data, rgb2gray

psf_fp = "data/psf/tape_rgb.png"
data_fp = "data/raw_data/thumbs_up_rgb.png"
downsample = 8


def test_load_data():
    for gray in [True, False]:
        for dtype in ["float32", "float64"]:
            psf, data = load_data(
                psf_fp=psf_fp,
                data_fp=data_fp,
                downsample=downsample,
                plot=False,
                gray=gray,
                dtype=dtype,
            )
            assert psf.shape[3] == (1 if gray else 3)
            assert len(psf.shape) == 4
            assert len(data.shape) == 4
            assert data.shape[0] == 1
            assert psf.shape[1:] == data.shape[1:]
            assert psf.dtype == dtype, dtype
            assert data.dtype == dtype, dtype


def test_rgb2gray():
    for is_torch in [True, False]:
        psf, data = load_data(
            psf_fp=psf_fp,
            data_fp=data_fp,
            downsample=downsample,
            plot=False,
            dtype="float32",
            use_torch=is_torch,
        )
        data = data[0]  # drop first depth dimension

        # try with 4D
        psf_gray = rgb2gray(psf, keepchanneldim=False)
        assert len(psf_gray.shape) == 3
        psf_gray = rgb2gray(psf, keepchanneldim=True)
        assert len(psf_gray.shape) == 4

        # try with 3D
        data_gray = rgb2gray(data, keepchanneldim=False)
        assert len(data_gray.shape) == 2
        data_gray = rgb2gray(data, keepchanneldim=True)
        assert len(data_gray.shape) == 3


if __name__ == "__main__":
    test_load_data()
    test_rgb2gray()



# #############################################################################
# sensor.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Sensor
======

This module provides utilities to simulate a camera sensor and obtain its specifications
(resolution, pixel size, diagonal).

On the roadmap:

* Couple :py:class:`~lensless.sensor.VirtualSensor` with lens / mask / aperture.
* Add noise to :py:class:`~lensless.sensor.VirtualSensor`.
* Set object height :py:class:`~lensless.sensor.VirtualSensor`.
* Similar class as :py:class:`~lensless.sensor.VirtualSensor` for "real" sensors to capture images.

"""


import numpy as np
from enum import Enum
from cv2 import resize

from lensless.utils.image import rgb2gray
from lensless.utils.io import load_image


class SensorOptions(Enum):
    """
    Available sensor options.

    * ``rpi_hq``: `Raspberry Pi HQ Camera Sensor (IMX 477) <https://www.raspberrypi.com/products/raspberry-pi-high-quality-camera/>`_
    * ``rpi_gs``: `Raspberry Pi Global Shutter Camera Sensor (IMX 296) <https://www.raspberrypi.com/products/raspberry-pi-global-shutter-camera/>`_
    * ``rpi_v2``: `Raspberry Pi Camera Module V2 Sensor (IMX 219) <https://www.raspberrypi.com/products/camera-module-v2/>`_
    * ``basler_287``: `Basler daA720-520um Sensor (IMX 287) <https://www.baslerweb.com/en/products/cameras/area-scan-cameras/dart/daa720-520um-cs-mount/>`_
    * ``basler_548``: `Basler daA2448-70uc Sensor (IMX 548) <https://www.baslerweb.com/en/products/cameras/area-scan-cameras/dart/daa2448-70uc-cs-mount/>`_
    """

    RPI_HQ = "rpi_hq"
    RPI_GS = "rpi_gs"
    RPI_V2 = "rpi_v2"
    BASLER_287 = "basler_287"
    BASLER_548 = "basler_548"

    @staticmethod
    def values():
        return [dev.value for dev in SensorOptions]


class SensorParam:
    PIXEL_SIZE = "pixel_size"
    RESOLUTION = "resolution"
    DIAGONAL = "diagonal"
    COLOR = "color"
    BIT_DEPTH = "bit_depth"
    MAX_EXPOSURE = "max_exposure"  # in seconds
    MIN_EXPOSURE = "min_exposure"  # in seconds


"""
Note sensors are in landscape orientation.

Max exposure for RPi cameras: https://www.raspberrypi.com/documentation/accessories/camera.html#hardware-specification
"""
sensor_dict = {
    # Raspberry Pi HQ Camera Sensor
    # datasheet: https://www.arducam.com/sony/imx477/#imx477-datasheet
    # IMX 477
    SensorOptions.RPI_HQ.value: {
        SensorParam.PIXEL_SIZE: np.array([1.55e-6, 1.55e-6]),
        SensorParam.RESOLUTION: np.array([3040, 4056]),
        SensorParam.DIAGONAL: 7.857e-3,
        SensorParam.COLOR: True,
        SensorParam.BIT_DEPTH: [8, 12],
        SensorParam.MAX_EXPOSURE: 670.74,
        SensorParam.MIN_EXPOSURE: 0.02,
    },
    # Raspberry Pi Global Shutter Camera
    # https://www.raspberrypi.com/products/raspberry-pi-global-shutter-camera/
    # IMX 296
    SensorOptions.RPI_GS.value: {
        SensorParam.PIXEL_SIZE: np.array([3.45e-6, 3.45e-6]),
        SensorParam.RESOLUTION: np.array([1088, 1456]),
        SensorParam.DIAGONAL: 6.3e-3,
        SensorParam.COLOR: True,
        SensorParam.BIT_DEPTH: [8, 10],
        SensorParam.MAX_EXPOSURE: 15534385e-6,
        SensorParam.MIN_EXPOSURE: 29e-6,
    },
    # Raspberry Pi Camera Module V2
    # https://www.raspberrypi.com/documentation/accessories/camera.html#hardware-specification
    # IMX 219
    SensorOptions.RPI_V2.value: {
        SensorParam.PIXEL_SIZE: np.array([1.12e-6, 1.12e-6]),
        SensorParam.RESOLUTION: np.array([2464, 3280]),
        SensorParam.DIAGONAL: 4.6e-3,
        SensorParam.COLOR: True,
        SensorParam.BIT_DEPTH: [8],
        SensorParam.MAX_EXPOSURE: 11.76,
        SensorParam.MIN_EXPOSURE: 0.02,  # TODO : verify
    },
    # Basler daA720-520um
    # https://www.baslerweb.com/en/products/cameras/area-scan-cameras/dart/daa720-520um-cs-mount/
    # IMX 287
    SensorOptions.BASLER_287.value: {
        SensorParam.PIXEL_SIZE: np.array([6.9e-6, 6.9e-6]),
        SensorParam.RESOLUTION: np.array([540, 720]),
        # SensorParam.DIAGONAL: ,
        SensorParam.COLOR: False,
        SensorParam.BIT_DEPTH: [8, 12],
    },
    # Basler daA2448-70uc
    # https://www.baslerweb.com/en/products/cameras/area-scan-cameras/dart/daa2448-70uc-cs-mount/
    # IMX 548
    SensorOptions.BASLER_548.value: {
        SensorParam.PIXEL_SIZE: np.array([2.74e-6, 2.74e-6]),
        SensorParam.RESOLUTION: np.array([2048, 2448]),
        # 8.8 in other sources: https://www.gophotonics.com/products/cmos-image-sensors/sony-corporation/21-209-imx548?utm_source=gophotonics&utm_medium=similar
        # 8.7 in Basler docs but leads to error in shape...
        SensorParam.DIAGONAL: 8.8e-3,
        SensorParam.COLOR: True,
        SensorParam.BIT_DEPTH: [8, 10, 12],
    },
}


class VirtualSensor(object):
    """
    Virtual sensor class to simulate capturing a scene.
    """

    def __init__(
        self,
        pixel_size,
        resolution,
        diagonal=None,
        color=True,
        bit_depth=None,
        downsample=None,
        **kwargs,
    ):
        """
        Base constructor.

        Parameters
        ----------
        pixel_size : array-like or float
            2D pixel size in meters.
        resolution : array-like
            2D resolution in pixels.
        diagonal : float, optional
            Diagonal size in meters.
        color : bool, optional
            Whether the sensor is color or monochrome.
        bit_depth : list, optional
            List of supported bit depths.
        downsample : int, optional
            Downsample the sensor by this factor. Pixel size and resolution are adjusted accordingly.

        """

        assert len(resolution) == 2, "Resolution must be 2D"
        self.resolution = (
            resolution.copy()
        )  # to not overwrite original values when using downsample

        if isinstance(pixel_size, float):
            pixel_size = np.array([pixel_size, pixel_size])
        assert len(pixel_size) == 2, "Pixel size must be 2D"
        self.pixel_size = pixel_size.copy()

        self.diagonal = diagonal
        self.color = color
        if bit_depth is None:
            self.bit_depth = [8]
        else:
            self.bit_depth = bit_depth

        if diagonal is not None:
            # account for possible deadspace
            self.size = self.diagonal / np.linalg.norm(self.resolution) * self.resolution
        else:
            self.size = self.pixel_size * self.resolution

        self.pitch = self.size / self.resolution

        self.image_shape = self.resolution
        if self.color:
            self.image_shape = np.append(self.image_shape, 3)

        if downsample is not None:
            self.downsample(downsample)

    # contructor from sensor name
    @classmethod
    def from_name(cls, name, downsample=None):
        """
        Create a sensor from one of the available options in :py:class:`~lensless.hardware.sensor.SensorOptions`.

        Parameters
        ----------
        name : str
            Name of the sensor.

        Returns
        -------
        sensor : :py:class:`~lensless.sensor.VirtualSensor`
            Sensor.

        """

        if name not in SensorOptions.values():
            raise ValueError(f"Sensor {name} not supported.")
        sensor_specs = sensor_dict[name].copy()
        return cls(**sensor_specs, downsample=downsample)

    def capture(self, scene=None, bit_depth=None, bayer=False):
        """
        Virtual capture of a scene (assuming perfectly focused lens).

        Parameters
        ----------
        scene : :py:class:`~numpy.ndarray`, str, optional
            Scene to capture.
        bit_depth : int, optional
            Bit depth of the image. By default, use first available.
        bayer : bool, optional
            Whether to return a Bayer image or not. By default, return RGB (if color).

        Returns
        -------
        img : :py:class:`~numpy.ndarray`
            Captured image.

        """

        if bayer:
            raise NotImplementedError("Bayer capture not implemented yet.")

        if scene is None:
            scene = np.random.rand(*self.image_shape)
        else:

            if isinstance(scene, str):
                scene = load_image(scene)
            else:
                # check provided data has good shape
                if len(scene.shape) == 3:
                    assert scene.shape[2] == 3
                else:
                    assert len(scene.shape) == 2

            # rescale and keep aspect ratio
            scale = np.min(np.array(self.resolution) / np.array(scene.shape[:2]))
            dsize = tuple((np.array(scene.shape[:2]) * scale).astype(int))
            scene = resize(scene, dsize=dsize[::-1])
            diff = np.array(self.resolution) - np.array(scene.shape[:2])

            # pad if necessary
            if np.any(diff):

                # center padding
                pad_width = (
                    (diff[0] // 2, diff[0] - diff[0] // 2),
                    (diff[1] // 2, diff[1] - diff[1] // 2),
                )
                if len(scene.shape) == 3:
                    pad_width = pad_width + ((0, 0),)
                scene = np.pad(scene, pad_width, mode="constant", constant_values=0)

        assert scene is not None

        # convert to grayscale if necessary
        if not self.color:
            if len(scene.shape) == 3:
                scene = rgb2gray(scene, keepchanneldim=False)

        else:
            if len(scene.shape) == 2:
                # repeat channels
                scene = np.repeat(scene[:, :, np.newaxis], 3, axis=2)

        # normalize
        scene = scene.astype(np.float32)
        scene /= scene.max()

        # cast to appropriate bit depth
        if bit_depth is None:
            bit_depth = self.bit_depth[0]
        else:
            if bit_depth not in self.bit_depth:
                raise ValueError(f"Bit depth {bit_depth} not supported.")

        scene = (2**bit_depth - 1) * scene
        if bit_depth == 8:
            scene = scene.astype(np.uint8)
        elif bit_depth > 8:
            scene = scene.astype(np.uint16)

        return scene

    def downsample(self, factor):
        """
        Downsample the sensor by a given factor. Pixel size and resolution are adjusted accordingly.

        Parameters
        ----------
        factor : int
            Downsample factor.

        """

        assert factor > 1, "Downsample factor must be greater than 1."

        self.pixel_size = self.pixel_size * factor
        self.pitch = self.pitch * factor
        self.resolution = (self.resolution / factor).astype(int)
        self.size = self.pixel_size * self.resolution
        self.image_shape = self.resolution
        if self.color:
            self.image_shape = np.append(self.image_shape, 3)



"""
This script is used to export the .mat paf from https://github.com/Waller-Lab/DiffuserCam/tree/master/example_data
"""

import sys
import numpy as np
import scipy.io as sp


if len(sys.argv) < 2:
    print("Error : no filename provided. Aborting.")
    sys.exit()

filename = sys.argv[1]
if not filename.endswith(".mat"):
    print("Error : file is not a .mat file. Aborting")
    sys.exit()

img = np.array(list(sp.loadmat(filename).values()), dtype=object)[3]

img = img - np.min(img)
img = img / np.max(img)

img = np.swapaxes(img, 0, -1)
img = np.swapaxes(img, 1, 2)

np.save("psf.npy", img)



# Configuration file for the Sphinx documentation builder.
import datetime
import os
import sys
from unittest import mock


MOCK_MODULES = [
    "scipy",
    "scipy.signal",
    "scipy.linalg",
    "pycsou",
    "matplotlib",
    "matplotlib.pyplot",
    "scipy.fftpack",
    "cv2",
    "rawpy",
    "skimage.metrics",
    "lpips",
    "torchmetrics",
    "torchmetrics.image",
    "scipy.ndimage",
    "pycsou.abc",
    "pycsou.operator",
    "pycsou.operator.func",
    "pycsou.operator.linop",
    "pycsou.opt.solver",
    "pycsou.opt.stop",
    "pycsou.runtime",
    "pycsou.util",
    "pycsou.util.ptype",
    "PIL",
    "PIL.Image",
    "slm_controller",
    "slm_controller.hardware",
    "paramiko",
    "paramiko.ssh_exception",
    "perlin_numpy",
    "hydra",
    "hydra.utils",
    "scipy.special",
    "matplotlib.cm",
    "pyffs",
    "datasets",
    "huggingface_hub",
    "cadquery",
    "wandb",
    "einops",
]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = mock.Mock()
# -- Project information

sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))
from lensless import __version__

project = "LenslessPiCam"
copyright = f"{datetime.date.today().year}, Eric Bezzam"
author = "Eric Bezzam"
version = __version__
release = version

# -- General configuration

extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "NumPy [latest]": ("https://docs.scipy.org/doc/numpy/", None),
    "SciPy [latest]": ("https://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("http://matplotlib.sourceforge.net/", None),
    "pycsou": ("https://matthieumeo.github.io/pycsou/html/", None),
    "torch": ("https://pytorch.org/docs/stable/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for HTML output

html_theme = "sphinx_rtd_theme"
# html_theme_options = {"navigation_depth": -1, "titles_only": True}

# -- Options for napoleon extension ------------------------------------------
napoleon_google_docstring = False
napoleon_numpy_docstring = True



# #############################################################################
# fabrication.py
# =================
# Authors :
# Rein BENTDAL [rein.bent@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


"""
Mask Fabrication
================

This module provides tools for fabricating masks for lensless imaging.
Check out `this notebook <https://colab.research.google.com/drive/1eDLnDL5q4i41xPZLn73wKcKpZksfkkIo?usp=drive_link>`_ on Google Colab for how to use this module.

"""

import os
import cadquery as cq
import numpy as np
from typing import Union, Optional
from abc import ABC, abstractmethod
from lensless.hardware.mask import Mask, MultiLensArray, CodedAperture, FresnelZoneAperture


class Frame(ABC):
    @abstractmethod
    def generate(self, mask_size, depth: float) -> cq.Workplane:
        pass


class Connection(ABC):
    @abstractmethod
    def generate(self, mask: np.ndarray, mask_size, depth: float) -> cq.Workplane:
        """connections can in general use the mask array to determine where to connect to the mask, but it is not required."""
        pass


class Mask3DModel:
    def __init__(
        self,
        mask_array: np.ndarray,
        mask_size: Union[tuple[float, float], np.ndarray],
        height: Optional[float] = None,
        frame: Optional[Frame] = None,
        connection: Optional[Connection] = None,
        simplify: bool = False,
        show_axis: bool = False,
        generate: bool = True,
    ):
        """
        Wrapper to CadQuery to generate a 3D model from a mask array, e.g. for 3D printing.

        Parameters
        ----------
        mask_array : np.ndarray
            Array of the mask to generate from. 1 is opaque, 0 is transparent.
        mask_size : Union[tuple[float, float], np.ndarray]
            Dimensions of the mask in meters.
        height : Optional[float], optional
            How thick to make the mask in millimeters.
        frame : Optional[Frame], optional
            Frame object defining the frame around the mask.
        connection : Optional[Connection], optional
            Connection object defining how to connect the frame to the mask.
        simplify : bool, optional
            Combines all objects in the model to a single object. Can result in a smaller 3d model file and faster post processing. But takes a considerable amount of more time to generate model. Defaults to False.
        show_axis : bool, optional
            Show axis for debug purposes. Defaults to False.
        generate : bool, optional
            Generate model on initialization. Defaults to True.

        """

        self.mask = mask_array
        self.frame: Frame = frame
        self.connections: Connection = connection

        if isinstance(mask_size, tuple):
            self.mask_size = np.array(mask_size) * 1e3
        else:
            self.mask_size = mask_size * 1e3

        self.height = height
        self.simplify = simplify
        self.show_axis = show_axis

        self.model = None

        if generate:
            self.generate_3d_model()

    @classmethod
    def from_mask(cls, mask: Mask, **kwargs):
        """
        Create a Mask3DModel from a Mask object.

        Parameters
        ----------
        mask : :py:class:`~lensless.hardware.mask.Mask`
            Mask object to generate from, e.g. :py:class:`~lensless.hardware.mask.CodedAperture` or :py:class:`~lensless.hardware.mask.FresnelZoneAperture`.
        """
        assert isinstance(mask, CodedAperture) or isinstance(
            mask, FresnelZoneAperture
        ), "Mask must be a CodedAperture or FresnelZoneAperture object."
        return cls(mask_array=mask.mask, mask_size=mask.size, **kwargs)

    @staticmethod
    def mask_to_points(mask: np.ndarray, px_size: Union[tuple[float, float], np.ndarray]):
        """
        Turns mask into 2D point coordinates.

        Parameters
        ----------
        mask : np.ndarray
            Mask array.
        px_size : Union[tuple[float, float], np.ndarray]
            Pixel size in meters.

        """
        is_3D = len(np.unique(mask)) > 2

        if is_3D:
            indices = np.argwhere(mask != 0)
            coordinates = (indices - np.array(mask.shape) / 2) * px_size
            heights = mask[indices[:, 0], indices[:, 1]]

        else:
            indices = np.argwhere(mask == 0)
            coordinates = (indices - np.array(mask.shape) / 2) * px_size
            heights = None
        return coordinates, heights

    def generate_3d_model(self):
        """
        Based on provided (1) mask, (2) frame, and (3) connection between frame and mask, generate a 3d model.
        """

        assert self.model is None, "Model already generated."

        model = cq.Workplane("XY")

        if self.frame is not None:
            frame_model = self.frame.generate(self.mask_size, self.height)
            model = model.add(frame_model)
        if self.connections is not None:
            connection_model = self.connections.generate(self.mask, self.mask_size, self.height)
            model = model.add(connection_model)

        px_size = self.mask_size / self.mask.shape
        points, heights = Mask3DModel.mask_to_points(self.mask, px_size)
        if len(points) != 0:
            if heights is None:
                assert self.height is not None, "height must be provided if mask is 2D."
                mask_model = (
                    cq.Workplane("XY")
                    .pushPoints(points)
                    .box(px_size[0], px_size[1], self.height, centered=False, combine=False)
                )
            else:
                mask_model = cq.Workplane("XY")
                for point, height in zip(points, heights):

                    box = (
                        cq.Workplane("XY")
                        .moveTo(point[0], point[1])
                        .box(
                            px_size[0],
                            px_size[1],
                            height * self.height,
                            centered=False,
                            combine=False,
                        )
                    )
                    mask_model = mask_model.add(box)

            if self.simplify:
                mask_model = mask_model.combine(glue=True)

            model = model.add(mask_model)

        if self.simplify:
            model = model.combine(glue=False)

        if self.show_axis:
            axis_thickness = 0.01
            axis_length = 20
            axis_test = (
                cq.Workplane("XY")
                .box(axis_thickness, axis_thickness, axis_length)
                .box(axis_thickness, axis_length, axis_thickness)
                .box(axis_length, axis_thickness, axis_thickness)
            )
            model = model.add(axis_test)

        self.model = model

    def save(self, fname):
        """
        Save the 3d model to a file.

        Parameters
        ----------
        fname : str
            File name to save the model to.
        """

        assert self.model is not None, "Model not generated yet."

        directory = os.path.dirname(fname)
        if directory and not os.path.exists(directory):
            print(
                f"Error: The directory {directory} does not exist! Failed to save CadQuery model."
            )
            return

        cq.exporters.export(self.model, fname)


class MultiLensMold:
    def __init__(
        self,
        sphere_locations: np.ndarray,
        sphere_radius: np.ndarray,
        mask_size: Union[tuple[float, float], np.ndarray],
        mold_size: tuple[int, int, int] = (0.4e-1, 0.4e-1, 3.0e-3),
        base_height_mm: Optional[float] = 0.5,
        frame: Optional[Frame] = None,
        simplify: bool = False,
        show_axis: bool = False,
    ):
        """
        Create a 3D model of a multi-lens array mold.

        Parameters
        ----------
        sphere_locations : np.ndarray
            Array of sphere locations in meters.
        sphere_radius : np.ndarray
            Array of sphere radii in meters.
        mask_size : Union[tuple[float, float], np.ndarray]
            Dimensions of the mask in meters.
        mold_size : tuple[int, int, int], optional
            Dimensions of the mold in meters. Defaults to (0.4e-1, 0.4e-1, 3.0e-3).
        base_height_mm : Optional[float], optional
            Height of the base in millimeters. Defaults to 0.5.
        frame : Optional[Frame], optional
            Frame object defining the frame around the mask.
        simplify : bool, optional
            Combines all objects in the model to a single object. Can result in a smaller 3d model file and faster post processing. But takes a considerable amount of more time to generate model. Defaults to False.
        show_axis : bool, optional
            Show axis for debug purposes. Defaults to False.
        """

        self.mask_size_mm = mask_size * 1e3
        self.mold_size_mm = np.array(mold_size) * 1e3
        self.simplify = simplify
        self.frame = frame
        self.show_axis = show_axis
        self.n_lens = len(sphere_radius)

        # check mold larger than mask
        assert np.all(self.mask_size_mm <= self.mold_size_mm[:2]), "Mold must be larger than mask."
        assert base_height_mm < self.mold_size_mm[2], "Base height must be less than mold height."

        # create 3D model of multi-lens array
        model = cq.Workplane("XY")
        base_model = cq.Workplane("XY").box(
            self.mask_size_mm[0], self.mask_size_mm[1], base_height_mm, centered=(True, True, False)
        )
        model = model.add(base_model)

        if self.frame is not None:
            frame_model = self.frame.generate(self.mask_size_mm, base_height_mm)
            model = model.add(frame_model)

        sphere_model = cq.Workplane("XY")
        for i in range(self.n_lens):
            loc_mm = sphere_locations[i] * 1e3
            # # center locations
            loc_mm[0] -= self.mask_size_mm[1] / 2
            loc_mm[1] -= self.mask_size_mm[0] / 2
            r_mm = sphere_radius[i] * 1e3
            sphere = cq.Workplane("XY").moveTo(loc_mm[1], loc_mm[0]).sphere(r_mm, angle1=0)
            sphere_model = sphere_model.add(sphere)

        # add indent for removing
        if self.frame is not None:
            mask_dim = self.frame.size
        else:
            mask_dim = self.mask_size_mm
        # indent = cq.Workplane("XY").moveTo(0, mask_dim[1] / 2).sphere(base_height_mm, angle1=0)
        # indent = indent.translate((0, 0, -base_height_mm))
        indent = (
            cq.Workplane("XY")
            .moveTo(0, mask_dim[1] / 2)
            .box(base_height_mm, base_height_mm, base_height_mm)
        )
        indent = indent.translate((0, 0, -base_height_mm / 2))
        sphere_model = sphere_model.add(indent)

        # add to base
        sphere_model = sphere_model.translate((0, 0, base_height_mm))
        model = model.add(sphere_model)

        if self.simplify:
            model = model.combine(glue=True)

        if self.show_axis:
            axis_thickness = 0.01
            axis_length = 20
            axis_test = (
                cq.Workplane("XY")
                .box(axis_thickness, axis_thickness, axis_length)
                .box(axis_thickness, axis_length, axis_thickness)
                .box(axis_length, axis_thickness, axis_thickness)
            )
            model = model.add(axis_test)

        self.mask = model

        # create mold
        mold = cq.Workplane("XY").box(
            self.mold_size_mm[0],
            self.mold_size_mm[1],
            self.mold_size_mm[2],
            centered=(True, True, False),
        )
        mold = mold.cut(model).rotate((0, 0, 0), (1, 0, 0), 180)

        self.mold = mold

    @classmethod
    def from_mask(cls, mask: Mask, **kwargs):
        """
        Create a Mask3DModel from a Mask object.

        Parameters
        ----------
        mask : :py:class:`~lensless.hardware.mask.MultiLensArray`
            Multi-lens array mask object.
        """
        assert isinstance(mask, MultiLensArray), "Mask must be a MultiLensArray object."
        return cls(
            sphere_locations=mask.loc, sphere_radius=mask.radius, mask_size=mask.size, **kwargs
        )

    def save(self, fname):
        assert self.mold is not None, "Model not generated yet."

        directory = os.path.dirname(fname)
        if directory and not os.path.exists(directory):
            print(
                f"Error: The directory {directory} does not exist! Failed to save CadQuery model."
            )
            return

        cq.exporters.export(self.mold, fname)


# --- from here, implementations of frames and connections ---


class SimpleFrame(Frame):
    def __init__(self, padding: float = 2, size: Optional[tuple[float, float]] = None):
        """
        Specify either padding or size. If size is specified, padding is ignored.

        All dimensions are in millimeters.

        Parameters
        ----------
        padding : float, optional
            padding around the mask. Defaults to 2mm.
        size : Optional[tuple[float, float]], optional
            Size of the frame in mm. Defaults to None.
        """
        self.padding = padding
        self.size = size

    def generate(self, mask_size, depth: float) -> cq.Workplane:
        width, height = mask_size[0], mask_size[1]
        size = (
            self.size
            if self.size is not None
            else (width + 2 * self.padding, height + 2 * self.padding)
        )
        return (
            cq.Workplane("XY")
            .box(size[0], size[1], depth, centered=(True, True, False))
            .rect(width, height)
            .cutThruAll()
        )


class CrossConnection(Connection):
    """Transverse cross connection"""

    def __init__(self, line_width: float = 0.1, mask_radius: float = None):
        self.line_width = line_width
        self.mask_radius = mask_radius

    def generate(self, mask: np.ndarray, mask_size, depth: float) -> cq.Workplane:
        width, height = mask_size[0], mask_size[1]
        model = (
            cq.Workplane("XY")
            .box(self.line_width, height, depth, centered=(True, True, False))
            .box(width, self.line_width, depth, centered=(True, True, True))
        )

        if self.mask_radius is not None:
            circle = cq.Workplane("XY").cylinder(
                depth, self.mask_radius, centered=(True, True, False)
            )
            model = model.cut(circle)

        return model


class SaltireConnection(Connection):
    """Diagonal cross connection"""

    def __init__(self, line_width: float = 0.1, mask_radius: float = None):
        self.line_width = line_width
        self.mask_radius = mask_radius

    def generate(self, mask: np.ndarray, mask_size, depth: float) -> cq.Workplane:
        width, height = mask_size[0], mask_size[1]
        width2, height2 = width / 2, height / 2
        lw = self.line_width / np.sqrt(2)
        model = (
            cq.Workplane("XY")
            .moveTo(-(width2 - lw), -height2)
            .lineTo(-width2, -height2)
            .lineTo(-width2, -(height2 - lw))
            .lineTo(width2 - lw, height2)
            .lineTo(width2, height2)
            .lineTo(width2, height2 - lw)
            .close()
            .extrude(depth)
            .moveTo(-(width2 - lw), height2)
            .lineTo(-width2, height2)
            .lineTo(-width2, height2 - lw)
            .lineTo(width2 - lw, -height2)
            .lineTo(width2, -height2)
            .lineTo(width2, -(height2 - lw))
            .close()
            .extrude(depth)
        )

        if self.mask_radius is not None:
            circle = cq.Workplane("XY").cylinder(
                depth, self.mask_radius, centered=(True, True, False)
            )
            model = model.cut(circle)

        return model


class ThreePointConnection(Connection):
    """
    Connection for free-floating components as in FresnelZoneAperture.
    """

    def __init__(self, line_width: float = 0.1, mask_radius: float = None):
        self.line_width = line_width
        self.mask_radius = mask_radius

    def generate(self, mask: np.ndarray, mask_size, depth: float) -> cq.Workplane:
        width, height = mask_size[0], mask_size[1]
        width2, height2 = width / 2, height / 2
        lw = self.line_width / np.sqrt(2)

        model = (
            cq.Workplane("XY")
            .box(width2, self.line_width, depth, centered=(False, True, False))
            .moveTo(-(width2 - lw), -height2)
            .lineTo(-width2, -height2)
            .lineTo(-width2, -(height2 - lw))
            .lineTo(-lw, 0)
            .lineTo(lw, 0)
            .close()
            .extrude(depth)
            .moveTo(-(width2 - lw), height2)
            .lineTo(-width2, height2)
            .lineTo(-width2, (height2 - lw))
            .lineTo(-lw, 0)
            .lineTo(lw, 0)
            .close()
            .extrude(depth)
        )

        if self.mask_radius is not None:
            circle = cq.Workplane("XY").cylinder(
                depth, self.mask_radius, centered=(True, True, False)
            )
            model = model.cut(circle)

        return model


class CodedApertureConnection(Connection):
    def __init__(self, joint_radius: float = 0.1):
        self.joint_radius = joint_radius

    def generate(self, mask: np.ndarray, mask_size, depth: float) -> cq.Workplane:
        x_lines = np.where(np.diff(mask[:, 0]) != 0)[0] + 1
        y_lines = np.where(np.diff(mask[0]) != 0)[0] + 1
        X, Y = np.meshgrid(x_lines, y_lines)
        point_idxs = np.vstack([X.ravel(), Y.ravel()]).T - np.array(mask.shape) / 2

        px_size = mask_size / mask.shape
        points = point_idxs * px_size

        model = (
            cq.Workplane("XY")
            .pushPoints(points)
            .cylinder(depth, self.joint_radius, centered=(True, True, False), combine=False)
        )

        return model


def create_mask_adapter(
    fp, mask_w, mask_h, mask_d, adapter_w=12.90, adapter_h=9.90, support_w=0.4, support_d=0.4
):
    """
    Create and store an adapter for a mask given its measurements.
    Warning: Friction-fitted parts are to be made 0.05-0.1 mm smaller
    (ex: mask's width must fit in adapter's, adapter's width must fit in mount's, ...)

    Parameters
    ----------
    fp : string
        Folder in which to store the generated stl file.
    mask_w : float
        Length of the mask's width in mm.
    mask_h : float
        Length of the mask's height in mm.
    mask_d : float
        Thickness of the mask in mm.
    adapter_w : float
        Length of the adapter's width in mm.
        default: current mount dim (13 - 0.1 mm)
    adapter_h : float
        Length of the adapter's height in mm.
        default: current mount dim (1.5 - 0.1 mm)
    support_w : float
        Width of the small extrusion to support the mask in mm
        default : current mount's dim (10 - 0.1 mm)
    support_d : float
        Thickness of the small extrusion to support the mask in mm
    """
    epsilon = 0.2

    # Make sure the dimension are realistic
    assert mask_w < adapter_w - epsilon, "mask's width too big"
    assert mask_h < adapter_h - epsilon, "mask's height too big"
    assert mask_w - 2 * support_w > epsilon, "mask's support too big"
    assert mask_h - 2 * support_w > epsilon, "mask's support too big"
    assert os.path.exists(fp), "folder does not exist"

    file_name = os.path.join(fp, "mask_adapter.stl")

    # Prevent accidental overwrite
    if os.path.isfile(file_name):
        print("Warning: already find mask_adapter.stl at " + fp)
        if input("Overwrite ? y/n") != "y":
            print("Abort adapter generation.")
            return

    # Construct the outer layer of the mask
    adapter = (
        cq.Workplane("front")
        .rect(adapter_w, adapter_h)
        .rect(mask_w, mask_h)
        .extrude(mask_d + support_d)
    )

    # Construct the dent to keep the mask secure
    support = (
        cq.Workplane("front")
        .rect(mask_w, mask_h)
        .rect(mask_w - 2 * support_w, mask_h - 2 * support_w)
        .extrude(support_d)
    )

    # Join the 2 shape in one
    adapter = adapter.union(support)

    # Save into path
    cq.exporters.export(adapter, file_name)



from huggingface_hub import hf_hub_download
from lensless.utils.io import load_psf, save_image
from lensless.utils.image import gamma_correction
import os
import numpy as np
import torch
from lensless.hardware.trainable_mask import AdafruitLCD

gamma = 1.8

## TapeCam
repo_id = "bezzam/TapeCam-Mirflickr-25K"
psf = "psf.png"
downsample = 8
flip_ud = False

# DigiCam-CelebA
repo_id = "bezzam/DigiCam-CelebA-26K"
psf = "psf_measured.png"
downsample = 8
gamma = 1.5
flip_ud = False

# # DigiCam-MirFlickr-25K
# repo_id = "bezzam/DigiCam-Mirflickr-SingleMask-25K"
# psf = "mask_pattern.npy"
# downsample = 8
# flip_ud = True

# # Multi Mask
# repo_id = "bezzam/DigiCam-Mirflickr-MultiMask-25K"
# psf = "masks/mask_4.npy"
# downsample = 8
# flip_ud = True


psf_fp = hf_hub_download(repo_id=repo_id, filename=psf, repo_type="dataset")

# load PSF
if psf.endswith(".npy"):
    mask_vals = np.load(psf_fp)
    mask = AdafruitLCD(
        initial_vals=torch.from_numpy(mask_vals.astype(np.float32)),
        sensor="rpi_hq",
        slm="adafruit",
        downsample=downsample,
        flipud=flip_ud,
        use_waveprop=True,
        scene2mask=0.3,
        mask2sensor=0.002,
        deadspace=True,
    )
    psf = mask.get_psf().detach().numpy()
else:
    psf = load_psf(psf_fp, downsample=downsample, flip_ud=flip_ud)

psf = psf / psf.max()
if gamma > 1:
    psf = gamma_correction(psf, gamma=gamma)

# save as viewable PNG
fn = os.path.basename(repo_id) + "_psf.png"
save_image(psf, fn)
print(f"Saved PSF as {fn}")



"""
After downloading the example files:
- Reconstrution: https://drive.switch.ch/index.php/s/NdgHlcDeHVDH5ww?path=%2Freconstruction
- Original: https://drive.switch.ch/index.php/s/NdgHlcDeHVDH5ww?path=%2Foriginal

And placing in (respectively):
- data/reconstruction
- data/original

The script can be run with:
```
python scripts/eval/compute_metrics_from_original.py
```

"""

import hydra
import os
from hydra.utils import to_absolute_path
import numpy as np
import matplotlib.pyplot as plt
from lensless.utils.plot import plot_image
from lensless.utils.io import load_image
from lensless.eval.metric import mse, psnr, ssim, lpips, extract
import matplotlib

font = {"family": "DejaVu Sans", "size": 18}
matplotlib.rc("font", **font)


@hydra.main(
    version_base=None, config_path="../../configs", config_name="compute_metrics_from_original"
)
def compute_metrics(config):
    recon = to_absolute_path(config.files.recon)
    original = to_absolute_path(config.files.original)
    vertical_crop = config.alignment.vertical_crop
    horizontal_crop = config.alignment.horizontal_crop
    rotation = config.alignment.rotation
    verbose = config.verbose

    # load estimate
    est = np.load(recon)
    if verbose:
        print("estimate shape", est.shape)

    # load original image
    img = load_image(original)
    img = img / img.max()

    # extract matching parts from estimate and original
    est, img_resize = extract(est, img, vertical_crop, horizontal_crop, rotation, verbose=verbose)

    _, ax = plt.subplots(ncols=2, nrows=1, figsize=(10, 5))
    plot_image(est, ax=ax[0])
    ax[0].set_title("Reconstruction")
    plot_image(img_resize, ax=ax[1])
    ax[1].set_title("Original")

    print("\nMSE", mse(img_resize, est))
    print("PSNR", psnr(img_resize, est))
    print("SSIM", ssim(img_resize, est))
    print("LPIPS", lpips(img_resize, est))

    plt.savefig("comparison.png")
    save = os.getcwd() + "/comparison.png"
    print(f"Save comparison to {save}")
    plt.show()


if __name__ == "__main__":
    compute_metrics()
